#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 15 16:43:31 2017

@author: ruess
"""
import pytest
from masci_tools.io.kkr_params import kkrparams
from masci_tools.io.common_functions import open_general
import tempfile
import os

# helper functions


def check_full_dict(p, p0):
    """
    helper function that compares full dictionary
    """
    from numpy import ndarray, array
    for key in [i[0] for i in p.get_set_values()]:
        v = p.get_value(key)
        v0 = p0.get_value(key)
        if not isinstance(v, list) and not isinstance(v, ndarray):
            if v != v0:
                print(key, v, v0)
            assert v == v0
        elif not isinstance(v[0], str):
            if abs(array(v) - array(v0)).max() >= 10**-14:
                print(key, abs(array(v) - array(v0)).max())
            assert abs(array(v) - array(v0)).max() < 10**-14
        else:
            if set(v) - set(v0) != set():
                print(key, set(v) - set(v0))
            assert set(v) - set(v0) == set()


# tests


class Test_create_and_set_keys(object):  # pylint: disable=missing-class-docstring

    def test_create_params_with_inital_values(self):
        p = kkrparams(RBASIS=[0, 0, 0], params_type='voronoi')
        assert isinstance(p, kkrparams)
        assert p.values['<RBASIS>'] == [0, 0, 0]

    def test_default_values(self):
        p = kkrparams()
        assert p.values['EMIN'] is None

    def test_set_single_value(self):
        p = kkrparams()
        p.set_value('EMIN', 2)
        assert p.values['EMIN'] == 2.
        assert p.values['EMAX'] is None

    def test_set_multiple_values(self):
        p = kkrparams()
        p.set_multiple_values(EMIN=1, EMAX=2)
        assert p.values['EMIN'] == 1.
        assert p.values['EMAX'] == 2.


class Test_capture_wrong_input(object):  # pylint: disable=missing-class-docstring

    def test_wrong_input_type(self):
        p = kkrparams()
        with pytest.raises(TypeError):
            p.set_value('EMIN', '2')

        with pytest.raises(TypeError):
            p.set_value('EMIN', False)

    def test_wrong_input_array_dimension(self):
        p = kkrparams()
        from numpy import array, sqrt
        bravais = array([[0.7071067812, -0.5, 0.0], [0.7071067812, 0.5, 0.0], [sqrt(2), 0.0, 0.866025404]])

        # atom positions in relative coordinates
        basis_vectors = []
        for iatom in range(6):
            tmp = array([0, 0, 0]) + iatom * array([0.5, 0.5, bravais[2, 2]])
            tmp[0] = tmp[0] % 1
            tmp[1] = tmp[1] % 1
            print(iatom, tmp)
            basis_vectors.append(tmp)
        basis_vectors = array(basis_vectors)
        p.set_value('INTERFACE', True)
        p.set_value('<RBLEFT>', array([[1, 1], [0, 1]]))

    def test_input_consistency_check_fail(self):

        p = kkrparams(ZATOM=29.,
                      LMAX=2,
                      NAEZ=1,
                      BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                      RMAX=7,
                      GMAX=65,
                      NSPIN=2,
                      RBASIS=[0, 0, 0],
                      ALATBASIS=1)
        p.set_value('LDAU_PARA', [1, 2])
        with pytest.raises(TypeError):
            p._check_input_consistency()

    def test_inconsistency_bulk_mode_bravais(self):
        p = kkrparams(LMAX=2,
                      NAEZ=1,
                      BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 0]],
                      NSPIN=2,
                      RBASIS=[0, 0, 0],
                      ALATBASIS=1,
                      RMAX=7,
                      GMAX=65,
                      ZATOM=29.)
        with pytest.raises(ValueError):
            p.fill_keywords_to_inputfile()


class Test_get_info(object):  # pylint: disable=missing-class-docstring

    def test_get_mandatory(self):
        p = kkrparams()
        manlist = p.get_all_mandatory()
        assert set(manlist) == set(
            ['LMAX', 'NAEZ', 'BRAVAIS', 'RMAX', 'GMAX', 'NSPIN', '<RBASIS>', 'ALATBASIS', '<ZATOM>'])

    def test_get_set_values(self):
        p = kkrparams()
        setlist = p.get_set_values()
        assert setlist == []

    def test_get_set_values2(self):
        from numpy import array
        p = kkrparams()
        p.set_multiple_values(EMIN=1, EMAX=2)
        setlist = p.get_set_values()
        assert set(array(setlist).flatten()) == set(array([['EMIN', 1.], ['EMAX', 2.]]).flatten())

    def test_get_description(self):
        p = kkrparams()
        desc = p.get_description('EMIN')
        assert desc == 'Accuracy, Valence energy contour: Lower value (in Ryd) for the energy contour'

    def test_get_type(self):
        p = kkrparams()
        tlist = p.get_type('BRAVAIS')
        assert tlist == [float, float, float, float, float, float, float, float, float]

    def test_is_mandatory(self):
        p = kkrparams()
        man = p.is_mandatory('EMAX')
        assert not man

    def test_get_value(self):
        p = kkrparams(LMAX=3)
        # check for KeyError if wrong key is checked
        with pytest.raises(KeyError):
            p.get_value('something_wrong')

        # check for returning unset value
        npol = p.get_value('NPOL')
        assert npol is None
        # check correct LMAX value
        lmax = p.get_value('LMAX')
        assert lmax == 3
        # check for returning lists for RUNOPT and TESTOPT
        runopt = p.get_value('RUNOPT')
        testopt = p.get_value('TESTOPT')
        assert runopt == []
        assert testopt == []
        p = kkrparams(TESTOPT=['test1', 'test2'], RUNOPT=['NEWSOSOL'])
        runopt = p.get_value('RUNOPT')
        testopt = p.get_value('TESTOPT')
        assert runopt == ['NEWSOSOL']
        assert set(testopt) == set(['test1', 'test2'])


class Test_fill_inputfile(object):
    """
    Tests checking writing an input file
    """

    def test_fill_inputfile_filehandle(self, file_regression):
        p = kkrparams(params_type='kkrimp')
        p.set_multiple_values(CALCORBITALMOMENT=0,
                              RUNFLAG='',
                              QBOUND=10**-7,
                              NSPIN=1,
                              TESTFLAG='',
                              NPAN_EQ=7,
                              CALCFORCE=0,
                              NPAN_LOGPANELFAC=2,
                              SPINORBIT=0,
                              ITDBRY=20,
                              NPAN_LOG=5,
                              INS=1,
                              ICST=2,
                              CALCJIJMAT=0,
                              NCHEB=10,
                              HFIELD=[0.00, 0],
                              BRYMIX=0.05,
                              KVREL=1,
                              IMIX=0,
                              RADIUS_MIN=-1,
                              NCOLL=0,
                              RADIUS_LOGPANELS=0.6,
                              MIXFAC=0.05,
                              SCFSTEPS=1,
                              XC='LDA-VWN')
        with tempfile.NamedTemporaryFile('r') as tmp:
            p.fill_keywords_to_inputfile(output=tmp.name)
            file_content = tmp.read()

        file_regression.check(file_content)

    def test_fill_inputfile_minimal_Voronoi(self, file_regression):
        p = kkrparams(ZATOM=29.,
                      LMAX=2,
                      NAEZ=1,
                      BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                      RCLUSTZ=1.5,
                      NSPIN=2,
                      RBASIS=[0, 0, 0],
                      ALATBASIS=1)
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as td:
            os.chdir(td)
            p.fill_keywords_to_inputfile(is_voro_calc=True)
            file_content = open('inputcard').read()
            os.chdir(cwd)

        file_regression.check(file_content)

    def test_fill_inputfile_KKR(self, file_regression):
        p = kkrparams(ZATOM=29.,
                      LMAX=2,
                      NAEZ=1,
                      BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                      RMAX=7,
                      GMAX=65,
                      RCLUSTZ=1.5,
                      NSPIN=2,
                      RBASIS=[0, 0, 0],
                      ALATBASIS=1)
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as td:
            os.chdir(td)
            p.fill_keywords_to_inputfile()
            file_content = open('inputcard').read().strip()
            os.chdir(cwd)

        file_regression.check(file_content)

    def test_fill_inputfile_empty_check(self):
        p = kkrparams(LMAX=2, NAEZ=1)

        with pytest.raises(ValueError):
            p.fill_keywords_to_inputfile()

    def test_fill_inputfile_all_keys(self):
        """Example filling all keys"""
        from numpy import array, sqrt

        alat = 5.416871386
        naez = 6
        bravais = array([[0.7071067812, -0.5, 0.0], [0.7071067812, 0.5, 0.0], [sqrt(2), 0.0, 0.866025404]])
        lmax = 2
        nspin = 2
        nucl_numbers = [0, 0, 26, 27, 26, 27, 0, 0]
        cpa_info = [naez + 2, [1., 1., 0.98, 0.02, 0.98, 0.02, 1., 1.], [1, 2, 3, 3, 4, 4, 5, 6]]
        npol = 4
        npt1, npt2, npt3 = 3, 10, 3
        tempr = 800
        basis_vectors = []
        for iatom in range(naez):
            tmp = array([0, 0, 0]) + iatom * array([0.5, 0.5, bravais[2, 2]])
            tmp[0] = tmp[0] % 1
            tmp[1] = tmp[1] % 1
            print(iatom, tmp)
            basis_vectors.append(tmp)
        basis_vectors = array(basis_vectors)
        natyp = cpa_info[0]
        cpa_conc = cpa_info[1]
        cpa_sites = cpa_info[2]
        ins = 1
        kshape = ins
        rmax, gmax = 7, 65
        rcls = 1.5
        bzdivide = [10, 10, 0]
        emin = -0.4
        p = kkrparams()
        p.set_multiple_values(ZATOM=nucl_numbers, RBASIS=basis_vectors, BRAVAIS=bravais, NAEZ=naez, ALATBASIS=alat)
        p.set_multiple_values(NSPIN=nspin, LMAX=lmax, NPOL=npol, NPT1=npt1, NPT2=npt2, NPT3=npt3, TEMPR=tempr)
        p.set_multiple_values(RMAX=rmax, GMAX=gmax)
        p.set_multiple_values(RCLUSTZ=rcls, BZDIVIDE=bzdivide, EMIN=emin)
        p.set_multiple_values(INS=ins, KSHAPE=kshape)
        p.set_multiple_values(INTERFACE=True,
                              NLBASIS=1,
                              NRBASIS=1,
                              ZPERIODL=array([-0.5, -0.5, -bravais[2, 2]]),
                              ZPERIODR=array([0.5, 0.5, bravais[2, 2]]),
                              RBLEFT=basis_vectors[0] + array([-0.5, -0.5, -bravais[2, 2]]),
                              RBRIGHT=basis_vectors[naez - 1] + array([0.5, 0.5, bravais[2, 2]]))
        p.set_value('LINIPOL', True)
        p.set_value('XINIPOL', [1 for i in range(natyp)])
        p.set_value('HFIELD', 0.02)
        p.set_value('NSTEPS', 1)
        p.set_value('IMIX', 0)
        p.set_value('STRMIX', 0.01)
        p.set_value('CARTESIAN', False)
        p.set_multiple_values(KAOEZR=1,
                              KAOEZL=1,
                              FPRADIUS=[-1 for i in range(natyp)],
                              RCLUSTXY=rcls,
                              TKSEMI=800,
                              EMAX=1,
                              NPOLSEMI=0,
                              N2SEMI=0,
                              N1SEMI=0,
                              N3SEMI=0,
                              FSEMICORE=0,
                              KVREL=1,
                              NCHEB=7,
                              VCONST=0,
                              SOCSCL=[1 for i in range(natyp)],
                              LAMBDA_XC=1,
                              FCM=20,
                              ITDBRY=20,
                              KREADLDAU=0,
                              RUNOPT=['LDAU', 'SEMICORE', 'IRGENDWAS FALSCHES'],
                              TESTOPT=[
                                  'TSTOPTX0', 'TSTOPTX1', 'TSTOPTX2', 'TSTOPTX3', 'TSTOPTX4', 'TSTOPTX5', 'TSTOPTX6',
                                  'TSTOPTX7', 'TSTOPTX8', 'TSTOPTXYZZZZZZ'
                              ],
                              QBOUND=10**-3,
                              NPAN_LOG=3,
                              NPAN_EQ=4,
                              CPAINFO=[10**-3, 20],
                              LLOYD=0,
                              EMUSEMI=0,
                              ICST=2,
                              TOLRDIF=0.01,
                              BRYMIX=0.01,
                              EBOTSEMI=0,
                              NRIGHTHO=10,
                              KEXCOR=2,
                              NLEFTHOS=10,
                              R_LOG=0.4,
                              LDAU_PARA=[1, 2, 0, 0, 0],
                              NAT_LDAU=0,
                              RMTREFL=2.3,
                              RMTREFR=2.3,
                              DELTAE=[10**-5, 0],
                              RMTREF=[2.3 for i in range(natyp)])
        p.set_value('<SHAPE>', [1 for i in range(natyp)])
        p.set_multiple_values(NATYP=natyp, SITE=cpa_sites)
        p.set_value('<CPA-CONC>', cpa_conc)
        p.set_value('FILES', ['output.pot', ''])
        with tempfile.NamedTemporaryFile() as tmp:
            p.fill_keywords_to_inputfile(is_voro_calc=True, output=tmp.name)

    def test_set_rmtcore(self):
        #test rmtcore
        from numpy import array
        from masci_tools.io.common_functions import search_string

        para_dict = dict([(u'INS', 0), (u'RCLUSTZ', 1.69), (u'LMAX', 2), (u'GMAX', 65.0),
                          (u'<RMTCORE>', [0.3535533906, 0.3535533906, 0.3535533906, 0.3535533906]), (u'RMAX', 7.0),
                          (u'NSPIN', 1)])
        zatom = array([47., 47., 47., 47.])
        alat = 7.8692316414074615
        natom = 4
        positions = array([[0., 0., 0.], [0., 0.5, 0.5], [0.5, 0., 0.5], [0.5, 0.5, 0.]])
        bravais = array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
        k = kkrparams(**para_dict)
        k.set_multiple_values(ZATOM=zatom, NAEZ=natom, ALATBASIS=alat, RBASIS=positions, BRAVAIS=bravais)
        with tempfile.NamedTemporaryFile('r') as tmp:
            k.fill_keywords_to_inputfile(output=tmp.name)
            txt = tmp.readlines()

        naez = int(txt[search_string('NAEZ', txt)].split()[-1])
        rmtcore = []
        l_offset = search_string('RMTCORE', txt)
        for iatom in range(naez):
            rmtcore_at = float(txt[l_offset + 1 + iatom].split()[-1])
            rmtcore.append(rmtcore_at)
        maxdiff = (max(abs(array(para_dict['<RMTCORE>']) - array(rmtcore))))
        assert maxdiff < 10**-6

    def test_set_kkrimp_params_full(self, file_regression):
        p = kkrparams(params_type='kkrimp')
        p.set_multiple_values(CALCORBITALMOMENT=0,
                              RUNFLAG='',
                              QBOUND=10**-7,
                              NSPIN=1,
                              TESTFLAG='',
                              NPAN_EQ=7,
                              CALCFORCE=0,
                              NPAN_LOGPANELFAC=2,
                              SPINORBIT=0,
                              ITDBRY=20,
                              NPAN_LOG=5,
                              INS=1,
                              ICST=2,
                              CALCJIJMAT=0,
                              NCHEB=10,
                              HFIELD=[0.00, 0],
                              BRYMIX=0.05,
                              KVREL=1,
                              IMIX=0,
                              RADIUS_MIN=-1,
                              NCOLL=0,
                              RADIUS_LOGPANELS=0.6,
                              MIXFAC=0.05,
                              SCFSTEPS=1,
                              XC='LDA-VWN')
        with tempfile.NamedTemporaryFile('r') as tmp:
            p.fill_keywords_to_inputfile(output=tmp.name)
            file_content = tmp.read()

        file_regression.check(file_content)


class Test_read_inputfile(object):  # pylint: disable=missing-class-docstring

    def test_read_minimal_inputfile(self):
        p = kkrparams(ZATOM=26.,
                      LMAX=2,
                      NAEZ=1,
                      BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                      RCLUSTZ=1.5,
                      NSPIN=2,
                      RBASIS=[0, 0, 0],
                      ALATBASIS=1)
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as td:
            os.chdir(td)
            p.fill_keywords_to_inputfile(is_voro_calc=True)
            p2 = kkrparams(params_type='voronoi')
            p2.read_keywords_from_inputcard()
            os.chdir(cwd)

        check_full_dict(p, p2)

    def test_read_unsorted_inputfile(self):
        p = kkrparams(ZATOM=26.,
                      LMAX=2,
                      NAEZ=1,
                      BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                      RCLUSTZ=1.5,
                      NSPIN=2,
                      RBASIS=[0, 0, 0],
                      ALATBASIS=1,
                      RMAX=7,
                      GMAX=65)
        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as td:
            os.chdir(td)
            p.fill_keywords_to_inputfile(output='input.temp.txt')
            txt = open('input.temp.txt', 'r').readlines()
            # exchange some lines
            tmp = txt[0]
            txt[0] = txt[5]
            txt[5] = tmp
            tmp = txt[-1]
            txt[-1] = txt[-2]
            txt[-2] = tmp
            tmp = txt[-2]
            txt[-2] = txt[-4]
            txt[-4] = tmp
            tmp = txt[-3]
            txt[-3] = txt[-1]
            txt[-1] = tmp
            with open('input.temp_unsorted.txt', 'w') as f:
                f.writelines(txt)
            p2 = kkrparams()
            p2.read_keywords_from_inputcard(inputcard='input.temp_unsorted.txt')
            print(p2.get_dict())
            print(dict(p2.get_set_values()))
            check_full_dict(p, p2)
            os.chdir(cwd)

    def test_read_slab(self):
        from numpy import array
        from masci_tools.util.constants import BOHR_A
        p = kkrparams(params_type='kkr')

        # automatically read keywords from inpucard
        p.read_keywords_from_inputcard(inputcard='../tests/files/kkr/import_calc_old_style/inputcard')
        # convert some read-in stuff back from Ang. units to alat units
        rbl = p.get_value('<RBLEFT>')
        rbr = p.get_value('<RBRIGHT>')
        zper_l = p.get_value('ZPERIODL')
        zper_r = p.get_value('ZPERIODR')
        ang2alat = 1 / (p.get_value('ALATBASIS') * BOHR_A)
        if rbl is not None:
            p.set_value('<RBLEFT>', array(rbl) * ang2alat)
        if rbr is not None:
            p.set_value('<RBRIGHT>', array(rbr) * ang2alat)
        if zper_l is not None:
            p.set_value('ZPERIODL', array(zper_l) * ang2alat)
        if zper_r is not None:
            p.set_value('ZPERIODR', array(zper_r) * ang2alat)

        # set parameters of expected values manually
        p0 = kkrparams(RUNOPT=['xigid-ef', 'LLOYD', 'ewald2d', 'NEWSOSOL', 'DOS'],
                       TESTOPT=['ie', 'RMESH', 'clusters', 'MPIenerg', 'fullBZ', 'DOS'],
                       LMAX=3,
                       NSPIN=2,
                       NATYP=80,
                       NAEZ=80,
                       CARTESIAN=True,
                       ALATBASIS=20.156973053,
                       BRAVAIS=[[0.38437499, 0., 0.], [0.19218749, -0.33287851, 0.], [0.19218749, -0.11095950, 1.]],
                       INTERFACE=True,
                       NRIGHTHO=10,
                       NLEFTHOS=10,
                       NLBASIS=10,
                       NRBASIS=10,
                       ZPERIODL=[-1.92187500000000e-01, 1.10959504859881e-01, -1.000000000000e+00],
                       ZPERIODR=[1.92187500000000e-01, -1.10959504859881e-01, 1.000000000000e+00],
                       RCLUSTZ=0.65,
                       RCLUSTXY=0.65,
                       EMIN=-1.2,
                       EMAX=1.2,
                       TEMPR=473.,
                       NPOL=7,
                       NPT1=7,
                       NPT2=40,
                       NPT3=6,
                       KSHAPE=2,
                       INS=1,
                       ICST=2,
                       KEXCOR=2,
                       HFIELD=0,
                       VCONST=0,
                       NPAN_LOG=17,
                       NPAN_EQ=7,
                       NCHEB=12,
                       R_LOG=0.8,
                       BZDIVIDE=[40, 40, 1],
                       NSTEPS=500,
                       IMIX=5,
                       STRMIX=0.02,
                       FCM=20.,
                       QBOUND=10**-7,
                       BRYMIX=0.02,
                       ITDBRY=30,
                       LINIPOL=False,
                       FILES=['potential', 'shapefun'],
                       RMAX=15.,
                       GMAX=900.)
        p0.set_value('<ZATOM>', [
            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0,
            52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0,
            0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 51.0, 0.0,
            52.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 51.0, 0.0, 52.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
            0.0, 0.0
        ])
        p0.set_value('<SHAPE>', [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5,
            6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10
        ])
        p0.set_multiple_values(KAOEZR=list(range(1, 11)),
                               KAOEZL=list(range(1, 11)),
                               KVREL=1,
                               RMTREFL=[
                                   2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000, 2.3562000, 2.4948000,
                                   2.2671000, 2.2671000, 2.5740000
                               ],
                               RMTREFR=[
                                   2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000, 2.3562000, 2.4948000,
                                   2.2671000, 2.2671000, 2.5740000
                               ])
        p0.set_multiple_values(RMTREF=[
            2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000, 2.3562000, 2.4948000, 2.2671000, 2.2671000,
            2.5740000, 2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000, 2.3562000, 2.4948000, 2.2671000,
            2.2671000, 2.5740000, 2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000, 2.3562000, 2.4948000,
            2.2671000, 2.2671000, 2.5740000, 2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000, 2.3562000,
            2.4948000, 2.2671000, 2.2671000, 2.5740000, 2.2671000, 2.2671000, 2.4948000, 2.3562000, 2.3562000,
            2.3562000, 2.4948000, 2.2671000, 2.2671000, 2.5740000, 2.2671000, 2.2671000, 2.4948000, 2.3562000,
            2.3562000, 2.3562000, 2.4948000, 2.2671000, 2.2671000, 2.5740000, 2.2671000, 2.2671000, 2.4948000,
            2.3562000, 2.3562000, 2.3562000, 2.4948000, 2.2671000, 2.2671000, 2.5740000, 2.2671000, 2.2671000,
            2.4948000, 2.3562000, 2.3562000, 2.3562000, 2.4948000, 2.2671000, 2.2671000, 2.5740000
        ])
        p0.set_multiple_values(RBLEFT=[[-1.92187500000000e-01, 1.10959504859881e-01, -1.000000000000e+00],
                                       [8.32667268468867e-17, 2.77555756156289e-17, -9.49500000000000e-01],
                                       [1.92187500000000e-01, -1.10959504859881e-01, -8.33000000000000e-01],
                                       [3.84375000000000e-01, -2.21919009719762e-01, -7.16500000000000e-01],
                                       [8.32667268468867e-17, 0.000000000000e+00, -6.33000000000000e-01],
                                       [1.92187500000000e-01, -1.10959504859881e-01, -5.49500000000000e-01],
                                       [3.84375000000000e-01, -2.21919009719762e-01, -4.33000000000000e-01],
                                       [2.77555756156289e-17, 1.38777878078145e-17, -3.16500000000000e-01],
                                       [1.92187500000000e-01, -1.10959504859881e-01, -2.66000000000000e-01],
                                       [3.84375000000000e-01, -2.21919009719762e-01, -1.33000000000000e-01]],
                               RBRIGHT=[[1.53750000000000e+00, -8.87676038879049e-01, 8.000000000000e+00],
                                        [1.72968750000000e+00, -9.98635543738930e-01, 8.05050000000000e+00],
                                        [1.92187500000000e+00, -1.10959504859881e+00, 8.16700000000000e+00],
                                        [2.11406250000000e+00, -1.22055455345869e+00, 8.28350000000000e+00],
                                        [1.72968750000000e+00, -9.98635543738930e-01, 8.36700000000000e+00],
                                        [1.92187500000000e+00, -1.10959504859881e+00, 8.45050000000000e+00],
                                        [2.11406250000000e+00, -1.22055455345869e+00, 8.56700000000000e+00],
                                        [1.72968750000000e+00, -9.98635543738930e-01, 8.68350000000000e+00],
                                        [1.92187500000000e+00, -1.10959504859881e+00, 8.73400000000000e+00],
                                        [2.11406250000000e+00, -1.22055455345869e+00, 8.86700000000000e+00]],
                               RBASIS=[[0.0, 0.0, 0.0], [0.1921875, -0.110959504859881, 0.0505000000000001],
                                       [0.384375, -0.221919009719762, 0.167], [0.5765625, -0.332878514579644, 0.2835],
                                       [0.1921875, -0.110959504859881, 0.367], [0.384375, -0.221919009719762, 0.4505],
                                       [0.5765625, -0.332878514579644, 0.567], [0.1921875, -0.110959504859881, 0.6835],
                                       [0.384375, -0.221919009719762, 0.734], [0.5765625, -0.332878514579644, 0.867],
                                       [0.1921875, -0.110959504859881, 1.0], [0.384375, -0.221919009719762, 1.0505],
                                       [0.5765625, -0.332878514579643, 1.167], [0.76875, -0.443838019439525, 1.2835],
                                       [0.384375, -0.221919009719762, 1.367], [0.5765625, -0.332878514579643, 1.4505],
                                       [0.76875, -0.443838019439525, 1.567], [0.384375, -0.221919009719762, 1.6835],
                                       [0.5765625, -0.332878514579643, 1.734], [0.76875, -0.443838019439525, 1.867],
                                       [0.384375, -0.221919009719762, 2.0], [0.5765625, -0.332878514579643, 2.0505],
                                       [0.76875, -0.443838019439525, 2.167], [0.9609375, -0.554797524299406, 2.2835],
                                       [0.5765625, -0.332878514579643, 2.367], [0.76875, -0.443838019439525, 2.4505],
                                       [0.9609375, -0.554797524299406, 2.567], [0.5765625, -0.332878514579643, 2.6835],
                                       [0.76875, -0.443838019439525, 2.734], [0.9609375, -0.554797524299406, 2.867],
                                       [0.5765625, -0.332878514579643, 3.0], [0.76875, -0.443838019439525, 3.0505],
                                       [0.9609375, -0.554797524299406, 3.167], [1.153125, -0.665757029159287, 3.2835],
                                       [0.76875, -0.443838019439525, 3.367], [0.9609375, -0.554797524299406, 3.4505],
                                       [1.153125, -0.665757029159287, 3.567], [0.76875, -0.443838019439525, 3.6835],
                                       [0.9609375, -0.554797524299406, 3.734], [1.153125, -0.665757029159287, 3.867],
                                       [0.76875, -0.443838019439525, 4.0], [0.9609375, -0.554797524299406, 4.0505],
                                       [1.153125, -0.665757029159287, 4.167], [1.3453125, -0.776716534019168, 4.2835],
                                       [0.9609375, -0.554797524299406, 4.367], [1.153125, -0.665757029159287, 4.4505],
                                       [1.3453125, -0.776716534019168, 4.567], [0.9609375, -0.554797524299406, 4.6835],
                                       [1.153125, -0.665757029159287, 4.734], [1.3453125, -0.776716534019168, 4.867],
                                       [0.9609375, -0.554797524299406, 5.0], [1.153125, -0.665757029159287, 5.0505],
                                       [1.3453125, -0.776716534019168, 5.167], [1.5375, -0.887676038879049, 5.2835],
                                       [1.153125, -0.665757029159287, 5.367], [1.3453125, -0.776716534019168, 5.4505],
                                       [1.5375, -0.887676038879049, 5.567], [1.153125, -0.665757029159287, 5.6835],
                                       [1.3453125, -0.776716534019168, 5.734], [1.5375, -0.887676038879049, 5.867],
                                       [1.153125, -0.665757029159287, 6.0], [1.3453125, -0.776716534019168, 6.0505],
                                       [1.5375, -0.887676038879049, 6.167], [1.7296875, -0.99863554373893, 6.2835],
                                       [1.3453125, -0.776716534019168, 6.367], [1.5375, -0.887676038879049, 6.4505],
                                       [1.7296875, -0.99863554373893, 6.567], [1.3453125, -0.776716534019168, 6.6835],
                                       [1.5375, -0.887676038879049, 6.734], [1.7296875, -0.99863554373893, 6.867],
                                       [1.3453125, -0.776716534019168, 7.0], [1.5375, -0.887676038879049, 7.0505],
                                       [1.7296875, -0.99863554373893, 7.167], [1.921875, -1.10959504859881, 7.2835],
                                       [1.5375, -0.887676038879049, 7.367], [1.7296875, -0.99863554373893, 7.4505],
                                       [1.921875, -1.10959504859881, 7.567], [1.5375, -0.887676038879049, 7.6835],
                                       [1.7296875, -0.99863554373893, 7.734], [1.921875, -1.10959504859881, 7.867]])
        p0.set_value('DECIFILES', ['fedeci.fp1', 'fedeci.fp2'])

        # check all values
        check_full_dict(p, p0)


class Test_other(object):  # pylint: disable=missing-class-docstring

    def test_get_missing_keys(self):
        p = kkrparams()
        missing = p.get_missing_keys()
        assert set(missing) == set(
            ['<ZATOM>', 'BRAVAIS', 'LMAX', 'GMAX', 'RMAX', 'NAEZ', '<RBASIS>', 'NSPIN', 'ALATBASIS'])
        missing = p.get_missing_keys(use_aiida=True)
        assert set(missing) == set(['LMAX', 'GMAX', 'RMAX', 'NSPIN'])

        p = kkrparams(params_type='voronoi', EMIN=-2, LMAX=3)
        missing = p.get_missing_keys()
        assert set(missing) == set(['<ZATOM>', 'BRAVAIS', 'RCLUSTZ', 'NAEZ', '<RBASIS>', 'NSPIN', 'ALATBASIS'])

    def test_set_value_None(self):
        p = kkrparams()
        p.set_value('EMIN', -1)
        assert p.values['EMIN'] == -1

        p.set_value('EMIN', None)
        assert p.values['EMIN'] == -1

        p.remove_value('EMIN')
        assert p.values['EMIN'] is None

    def test_set_potname_empty(self):
        from masci_tools.io.common_functions import search_string
        p = kkrparams()
        p.set_multiple_values(RMAX=1,
                              GMAX=1,
                              NSPIN=1,
                              RBASIS=[0, 0, 0],
                              LMAX=2,
                              RCLUSTZ=1.2,
                              NAEZ=1,
                              ZATOM=[0],
                              BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                              ALATBASIS=1,
                              FILES=['', 'shapenew'])
        with tempfile.NamedTemporaryFile('r') as tmp:
            p.fill_keywords_to_inputfile(output=tmp.name)
            txt = tmp.readlines()
        itmp = search_string('FILES', txt)
        potname = txt[itmp + 2].split()[0]
        shapename = txt[itmp + 4].split()[0]
        assert potname == 'potential'
        assert shapename == 'shapenew'

    def test_get_dict(self):
        d0 = {
            'POT_NS_CUTOFF': None,
            '<RMTCORE>': None,
            'ICST': None,
            '<RMTREF>': None,
            'N1SEMI': None,
            '<FPRADIUS>': None,
            '<NRBASIS>': None,
            '<SOCSCL>': None,
            'XINIPOL': None,
            'EMAX': None,
            '<RBLEFT>': None,
            'NLEFTHOS': None,
            '<ZATOM>': [0.0],
            'RCLUSTXY': None,
            'NPAN_EQ': None,
            '<RBRIGHT>': None,
            'BRAVAIS': [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            'INS': None,
            'NAT_LDAU': None,
            '<RMTREFR>': None,
            'ZPERIODL': None,
            'TESTOPT': None,
            'KEXCOR': None,
            '<TOLRDIF>': None,
            'TEMPR': None,
            'EBOTSEMI': None,
            'NATYP': None,
            'RUNOPT': None,
            'HFIELD': None,
            'NPOL': None,
            'RCLUSTZ': 1.2,
            'ZPERIODR': None,
            'N3SEMI': None,
            'LMAX': 2,
            'ITDBRY': None,
            '<KAOEZR>': None,
            '<LLOYD>': None,
            'STRMIX': None,
            'CPAINFO': None,
            'FCM': None,
            '<SHAPE>': None,
            'NPAN_LOG': None,
            'CARTESIAN': None,
            'FSEMICORE': None,
            'LAMBDA_XC': None,
            'GMAX': None,
            '<CPA-CONC>': None,
            'RMAX': None,
            'NCHEB': None,
            'EMIN': None,
            'NAEZ': 1,
            '<DELTAE>': None,
            'KREADLDAU': None,
            '<RBASIS>': [0, 0, 0],
            '<SITE>': None,
            'NPT2': None,
            'NPT3': None,
            'NPT1': None,
            'N2SEMI': None,
            'NPOLSEMI': None,
            '<RMTREFL>': None,
            'FILES': ['', 'shapenew'],
            'LDAU_PARA': None,
            'NSPIN': 1,
            'QBOUND': None,
            'NRIGHTHO': None,
            'KVREL': None,
            'TKSEMI': None,
            '<KAOEZL>': None,
            'NSTEPS': None,
            'KSHAPE': None,
            '<NLBASIS>': None,
            'LINIPOL': None,
            'BZDIVIDE': None,
            'INTERFACE': None,
            'BRYMIX': None,
            'EMUSEMI': None,
            'ALATBASIS': 1.0,
            'R_LOG': None,
            'IMIX': None,
            'VCONST': None,
            'JIJRAD': None,
            'JIJRADXY': None,
            'JIJSITEI': None,
            'JIJSITEJ': None,
            'NSHELD': None,
            'NMIN': None,
            'IEMXD': None,
            'IRID': None,
            'IPAND': None,
            'EFSET': None,
            '<AT_SCALE_BDG>': None,
            '<CALC_COMPLEX_BANDSTRUCTURE>': None,
            '<CALC_EXCHANGE_COUPLINGS>': None,
            '<CALC_EXCHANGE_COUPLINGS_ENERGY>': None,
            '<CALC_GF_EFERMI>': None,
            '<CALC_GMAT_LM_FULL>': None,
            '<CALC_WRONSKIAN>': None,
            '<CUSTOM_TESTSTRING>': None,
            '<DECOUPLE_SPINS_CHEBY>': None,
            '<DELTA_BDG>': None,
            '<DIRAC_SCALE_SPEEFOFLIGHT>': None,
            '<DISABLE_CHARGE_NEUTRALITY>': None,
            '<DISABLE_PRINT_SERIALNUMBER>': None,
            '<DISABLE_REFERENCE_SYSTEM>': None,
            '<DISABLE_TMAT_SRATRICK>': None,
            '<FIX_NONCO_ANGLES>': None,
            '<FORMATTED_FILE>': None,
            '<IMPURITY_OPERATOR_ONLY>': None,
            '<LAMBDA_BDG>': None,
            '<MODIFY_SOC_DIRAC>': None,
            '<MPI_SCHEME>': None,
            '<NEWVERSION_BDG>': None,
            '<NO_MADELUNG>': None,
            '<PRINT_GIJ>': None,
            '<PRINT_GMAT>': None,
            '<PRINT_ICKECK>': None,
            '<PRINT_KMESH>': None,
            '<PRINT_KPOINTS>': None,
            '<PRINT_PROGRAM_FLOW>': None,
            '<PRINT_RADIAL_MESH>': None,
            '<PRINT_REFPOT>': None,
            '<PRINT_TAU_STRUCTURE>': None,
            '<PRINT_TMAT>': None,
            '<RELAX_SPINANGLE_DIRAC>': None,
            '<SEARCH_EFERMI>': None,
            '<SET_CHEBY_NOSOC>': None,
            '<SET_CHEBY_NOSPEEDUP>': None,
            '<SET_EMPTY_SYSTEM>': None,
            '<SET_GMAT_TO_ZERO>': None,
            '<SET_KMESH_LARGE>': None,
            '<SET_KMESH_SMALL>': None,
            '<SET_TMAT_NOINVERSION>': None,
            '<SIMULATE_ASA>': None,
            '<SLOW_MIXING_EFERMI>': None,
            '<STOP_1A>': None,
            '<STOP_1B>': None,
            '<STOP_1C>': None,
            '<SYMMETRIZE_GMAT>': None,
            '<SYMMETRIZE_POTENTIAL_CUBIC>': None,
            '<SYMMETRIZE_POTENTIAL_MADELUNG>': None,
            '<TORQUE_OPERATOR_ONLYMT>': None,
            '<TORQUE_OPERATOR_ONLYSPH>': None,
            '<USE_BDG>': None,
            '<USE_BROYDEN_SPINMIX>': None,
            '<USE_CHEBYCHEV_SOLVER>': None,
            '<USE_COND_LB>': None,
            '<USE_CONT>': None,
            '<USE_DECIMATION>': None,
            '<USE_DECI_ONEBULK>': None,
            '<USE_EWALD_2D>': None,
            '<USE_FULL_BZ>': None,
            '<USE_LDAU>': None,
            '<USE_LLOYD>': None,
            '<USE_QDOS>': None,
            '<USE_READCPA>': None,
            '<USE_RIGID_EFERMI>': None,
            '<USE_SEMICORE>': None,
            '<USE_SPHERICAL_POTENTIAL_ONLY>': None,
            '<USE_VIRTUAL_ATOMS>': None,
            '<WRITE_ANGLES_ALLITER>': None,
            '<WRITE_BDG_TESTS>': None,
            '<WRITE_COMPLEX_QDOS>': None,
            '<WRITE_CPA_PROJECTION_FILE>': None,
            '<WRITE_DECI_POT>': None,
            '<WRITE_DECI_TMAT>': None,
            '<WRITE_DENSITY_ASCII>': None,
            '<WRITE_DOS>': None,
            '<WRITE_DOS_LM>': None,
            '<WRITE_ENERGY_MESH>': None,
            '<WRITE_GENERALIZED_POTENTIAL>': None,
            '<WRITE_GMAT_ASCII>': None,
            '<WRITE_GMAT_FILE>': None,
            '<WRITE_GMAT_PLAIN>': None,
            '<WRITE_GREEN_HOST>': None,
            '<WRITE_GREEN_IMP>': None,
            '<WRITE_GREF_FILE>': None,
            '<WRITE_KKRIMP_INPUT>': None,
            '<WRITE_KKRSUSC_INPUT>': None,
            '<WRITE_KPTS_FILE>': None,
            '<WRITE_LLOYD_CDOS_FILE>': None,
            '<WRITE_LLOYD_DGREF_FILE>': None,
            '<WRITE_LLOYD_DTMAT_FILE>': None,
            '<WRITE_LLOYD_FILE>': None,
            '<WRITE_LLOYD_G0TR_FILE>': None,
            '<WRITE_LLOYD_TRALPHA_FILE>': None,
            '<WRITE_MADELUNG_FILE>': None,
            '<WRITE_PKKR_INPUT>': None,
            '<WRITE_PKKR_OPERATORS>': None,
            '<WRITE_POTENTIAL_TESTS>': None,
            '<WRITE_RHO2NS>': None,
            '<WRITE_RHOQ_INPUT>': None,
            '<WRITE_TB_COUPLING>': None,
            '<WRITE_TMAT_FILE>': None,
            'DECIFILES': None,
            'IVSHIFT': None,
            'NPRINCD': None,
            'SPINMIXALPHA': None,
            'SPINMIXMEMLEN': None,
            'SPINMIXNSIMPLE': None,
            'SPINMIXQBOUND': None,
        }

        p = kkrparams()
        p.set_multiple_values(RMAX=1,
                              GMAX=1,
                              NSPIN=1,
                              RBASIS=[0, 0, 0],
                              LMAX=2,
                              RCLUSTZ=1.2,
                              NAEZ=1,
                              ZATOM=[0],
                              BRAVAIS=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                              ALATBASIS=1,
                              FILES=['', 'shapenew'])
        assert set(d0.keys()) == set(p.get_dict().keys())

        l0 = [
            '<SHAPE>', 'KSHAPE', 'ZPERIODL', '<NRBASIS>', '<NLBASIS>', '<RBASIS>', 'NAEZ', 'CARTESIAN', '<RBRIGHT>',
            '<RBLEFT>', 'INTERFACE', 'BRAVAIS', 'ALATBASIS', 'ZPERIODR'
        ]
        assert set(p.get_dict(group='lattice').keys()) == set(l0)

        l0 = ['ZPERIODL', '<NRBASIS>', '<NLBASIS>', '<RBRIGHT>', '<RBLEFT>', 'INTERFACE', 'ZPERIODR']
        assert set(l0) == set(p.get_dict(group='lattice', subgroup='2D mode').keys())

    def test_get_KKRcalc_parameter_defaults(self):
        d = kkrparams.get_KKRcalc_parameter_defaults()
        from masci_tools.io.kkr_params import __kkr_default_params__
        d0 = __kkr_default_params__
        assert d[0] == d0



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides the functionality to create/load the schema_dict for the
FleurInputSchema.xsd
"""
from .fleur_schema_parser_functions import *  #pylint: disable=unused-wildcard-import
from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict
from lxml import etree


def create_inpschema_dict(path):
    """
    Creates dictionary with information about the FleurInputSchema.xsd.
    The functions, whose results are added to the schema_dict and the corresponding keys
    are defined in schema_actions

    :param path: str path to the folder containing the FleurInputSchema.xsd file

    """

    #Add new functionality to this dictionary here
    schema_actions = {
        'root_tag': get_root_tag,
        'tag_paths': get_tag_paths,
        '_basic_types': get_basic_types,
        'attrib_types': extract_attribute_types,
        'simple_elements': get_basic_elements,
        'unique_attribs': get_unique_attribs,
        'unique_path_attribs': get_unique_path_attribs,
        'other_attribs': get_other_attribs,
        'omitt_contained_tags': get_omittable_tags,
        'tag_info': get_tag_info,
    }

    #print(f'processing: {path}/FleurInputSchema.xsd')
    xmlschema = etree.parse(path)
    xmlschema = clear_xml(xmlschema)

    namespaces = {'xsd': 'http://www.w3.org/2001/XMLSchema'}
    inp_version = str(xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0])

    schema_dict = {}
    schema_dict['inp_version'] = inp_version
    for key, action in schema_actions.items():
        schema_dict[key] = action(xmlschema, namespaces, **schema_dict)

    #We cannot do the conversion to CaseInsensitiveDict before since we need the correct case
    #For these attributes in the attrib_path functions
    schema_dict['simple_elements'] = CaseInsensitiveDict(schema_dict['simple_elements'])

    return schema_dict



# -*- coding: utf-8 -*-
"""
Tests of the inp.xml parser for Fleur
"""
import pytest
from masci_tools.io.parsers.fleur.fleur_inpxml_parser import inpxml_parser
import os

# Collect the input files
file_path1 = 'files/fleur/aiida_fleur/inpxml'
file_path2 = 'files/fleur/Max-R5'
file_path3 = 'files/fleur/aiida_fleur/nonvalid_inpxml'

inpxmlfilefolder = os.path.dirname(os.path.abspath(__file__))
inpxmlfilefolder_non_valid = [os.path.abspath(os.path.join(inpxmlfilefolder, file_path3))]

inpxmlfilefolder_valid = [
    os.path.abspath(os.path.join(inpxmlfilefolder, file_path1)),
    os.path.abspath(os.path.join(inpxmlfilefolder, file_path2))
]

#Thes inputs are currently broken in the fleur tests
broken_inputs = [
    'CoHybridPBE0', 'CoUnfold', 'gw1Interface', 'GaAsWannSOC', 'TiO2eelsXML', 'gw2Interface', 'Fe_film_SS_conv',
    'SiHybrid8kpt_nosym', 'Fe_bulk_SS_conv', 'Fe_film_SSFT', 'Max-R5/NiO_ldauXML', 'Max-R5/Bi2Te3XML'
]

inp_dict_input = ['FePt_film_SSFT_LO/files/inp2.xml']

inpxmlfilelist = []
inpxmlfilelist2 = []
inpxmlfilelist3 = []
for folder in inpxmlfilefolder_valid:
    for subdir, dirs, files in os.walk(folder):
        for file in files:
            if file.endswith('.xml') and 'inp' in file:
                non_valid = False
                for broken in broken_inputs:
                    if broken in subdir:
                        inpxmlfilelist2.append(os.path.join(subdir, file))
                        non_valid = True
                if not non_valid:
                    inpxmlfilelist.append(os.path.join(subdir, file))
                    for inp_dictfolder in inp_dict_input:
                        if inp_dictfolder in os.path.join(subdir, file):
                            inpxmlfilelist3.append(os.path.join(subdir, file))

for folder in inpxmlfilefolder_non_valid:
    for subdir, dirs, files in os.walk(folder):
        for file in files:
            if file.endswith('.xml') and 'inp' in file:
                inpxmlfilelist2.append(os.path.join(subdir, file))


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist)
def test_inpxml_valid_inpxml(inpxmlfilepath):
    """
    test if valid inp.xml files are recognized by the inpxml_parser
    """
    from lxml import etree

    #Pass inpxmlfile
    inp_dict = inpxml_parser(inpxmlfilepath)

    assert inp_dict is not None
    assert isinstance(inp_dict, dict)
    assert inp_dict != {}

    #Parse before
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(inpxmlfilepath, parser)
    inp_dict = inpxml_parser(xmltree)

    assert inp_dict is not None
    assert isinstance(inp_dict, dict)
    assert inp_dict != {}

    #Pass file handle
    with open(inpxmlfilepath, 'r') as inpfile:
        inp_dict = inpxml_parser(inpfile)

    assert inp_dict is not None
    assert isinstance(inp_dict, dict)
    assert inp_dict != {}


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist2)
def test_inpxml_non_valid_inpxml(inpxmlfilepath):
    """
    test if valid inp.xml files are recognized by the inpxml_parser
    """

    #Pass Path
    with pytest.raises((ValueError, FileNotFoundError)):
        inp_dict = inpxml_parser(inpxmlfilepath)


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist3, ids=['Fe_film_SSFT_LO'])
def test_inpxml_todict(data_regression, inpxmlfilepath):
    """
    test if valid inp.xml files are translated to the correct inp_dict
    """

    inp_dict = inpxml_parser(inpxmlfilepath)
    data_regression.check(inp_dict)


def test_inpxml_todict_warnings(data_regression, clean_parser_log):
    """
    test if valid inp.xml files are translated to the correct inp_dict
    """

    input_invalid_attr = os.path.abspath(os.path.join(inpxmlfilefolder, 'files/fleur/inp_invalid_attributes.xml'))
    warnings = {}

    #The parser shoul not raise and just log all the failed conversions
    inp_dict = inpxml_parser(input_invalid_attr, parser_info_out=warnings)
    data_regression.check({'input_dict': inp_dict, 'warnings': clean_parser_log(warnings)})


def test_inpxml_newer_version(data_regression, clean_parser_log):
    """
    test if valid inp.xml files with not yet existent versions are parsed correctly (fall back to latest available)
    """

    INPXML_FILEPATH = os.path.abspath(os.path.join(inpxmlfilefolder, 'files/fleur/input_newer_version.xml'))
    warnings = {}
    #The parser shoul not raise and just log all the failed conversions
    inp_dict = inpxml_parser(INPXML_FILEPATH, parser_info_out=warnings)
    data_regression.check({'input_dict': inp_dict, 'warnings': clean_parser_log(warnings)})



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""This module contains a simple class for set-like chemical elements enumeration."""
import dataclasses as dc

import masci_tools.vis.plot_methods


class ChemicalElements:
    def __init__(self, elements=None, empty=False, groups=None, distinct=False, special_elements: dict = None,
                 filepath=None):
        """A container for safe chemical element enumeration.

        Attribute 'elmts' stores either (flat) a dict key = chemical sumbol str : value = atomic number int.
        Or altternatively (nested), a dict of named group of such dicts (e.g. chemical property groups).
        The element dicts are always sorted by atomic number.
        If no or empty elements list/dict is supplied at initialization, will populate elmts with
        dict of entire periodic table.
        Class checks validity and prevents duplicate entries (within one group, not across groups).
        Class provides convenience methods: add/remove, set operations including overloaded operators
        '+' (union) and '-' (left difference), plot periodic table,
        container access methods: 'in' operator, []-operator for element (if flat) or group (if nested) access,
        relational operators (<, <=, ...).

        Additionally, 'data' and 'set_data()' arbitrary object storage for each group of chemical elements.

        DEVNOTES:
        - take care to choose correct 'elmts' access: elmts for interface, __elmts for internal.
        - take care to carry over self.distinct property to new CE objects

        :param elements: collection of element symbols or atomic numbers.
        :type elements: one of dict, list, tuple, set.
        :param empty: True: initialize without any elements.
        :param groups: If not None and more than one, ignore other arguments and initialize with empty groups.
        :type groups: list of strings.
        :param distinct: True: disallow same element in different groups.
        :param special_elements: dict symbol:atomic_number of special elements (eg {'X':0} for vacuum)
        :param filepath: if not None, init from JSON file. Must have been written with ChemicalElements.to_file().
        :type filepath: pathlib.Path

        >>> from masci_tools.util.chemical_elements import ChemicalElements
        >>> a = ChemicalElements([11, 2, 118, 78])
        >>> b = ChemicalElements(['Na', 'He', 'Og', 'Pt'])
        >>> c = ChemicalElements({'Na' : 11, 'He' : 2, 'Og' : 118, 'Pt' : 78})
        >>> d = ChemicalElements({'a' : ['Na', 'He'], 'b' : ['Og']})
        >>> d.add_elements(['Pt'], 'b')
        >>> e = ChemicalElements()
        >>> e.remove_elements([11, 2, 118, 78])
        >>> assert a.elmts == {'He': 2, 'Na': 11, 'Pt': 78, 'Og': 118}
        >>> assert a == b
        >>> assert a + b == c
        >>> assert a.intersection(b) == c
        >>> assert a.union(b) == ChemicalElements(d.flatten())
        >>> assert e.complement() == c
        """
        # property: distinctness
        self.distinct = distinct

        # periodic table, all elements; special element definitions not in actual periodic table
        # devnote: mendeleev.elements.get_all_elements() is very expensive (0.2s). hardcoded pte is faster.
        # self._pte = {el.symbol: el.atomic_number for el in mendeleev.elements.get_all_elements()}
        self._pte = {'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10, 'Na': 11,
                     'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20, 'Sc': 21,
                     'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30, 'Ga': 31,
                     'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40, 'Nb': 41,
                     'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50, 'Sb': 51,
                     'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60, 'Pm': 61,
                     'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70, 'Lu': 71,
                     'Hf': 72, 'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80, 'Tl': 81,
                     'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90, 'Pa': 91,
                     'U': 92, 'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100,
                     'Md': 101, 'No': 102, 'Lr': 103, 'Rf': 104, 'Db': 105, 'Sg': 106, 'Bh': 107, 'Hs': 108, 'Mt': 109,
                     'Ds': 110, 'Rg': 111, 'Cn': 112, 'Nh': 113, 'Fl': 114, 'Mc': 115, 'Lv': 116, 'Ts': 117, 'Og': 118}
        self._pte_inv = {v: k for k, v in self._pte.items()}

        self._special_elements = {}
        self._special_elements_inv = {}
        if special_elements:
            # assume correct format
            for symbol, number in special_elements.items():
                self.expand_allowed_elements(symbol=symbol, atomic_number=number)

                # if from file,
        if filepath:
            import json
            with open(filepath, 'r') as file:
                self.__elmts = json.load(file)
            self.__data = {group_name: None for group_name in self.__elmts.keys()}
        else:
            # init elmts, convert to [optional: container of:] dict[s] sym:num.
            import copy
            if groups:
                if not isinstance(groups, list) \
                        or not all([isinstance(group_name, str) for group_name in groups]):
                    raise ValueError("If groups not None, must be a list of strings.")
                if len(groups) < 2:
                    raise ValueError("If groups not None, must be more than one group. "
                                     "Else init with 'empty' or 'elements' instead.")

                self.__elmts = {group_name: {} for group_name in groups}
                self.__data = {group_name: None for group_name in groups}
            else:
                self.__elmts = {}
                if not empty:
                    list_types = (list, tuple, set)
                    if not elements:
                        # fill with whole periodic table
                        self.__elmts = {"": copy.deepcopy(self._pte)}

                    elif isinstance(elements, list_types):
                        # not nested
                        self.__elmts = {"": self._chemical_element_list_to_dict(elements)}
                    elif isinstance(elements, dict):
                        if not all([isinstance(v, list_types) for k, v in elements.items()]):
                            # flact dict
                            elements, _ = self._validate(elements)
                            # lazy type checking for sym:num
                            key_type_is_int = isinstance(list(elements.keys())[0], int)
                            elmts = {v: k for k, v in elements.items()} if key_type_is_int else elements
                            assert (all([isinstance(k, str) for k in elmts.keys()]))
                            assert (all([isinstance(v, int) for v in elmts.values()]))
                            self.__elmts = {"": self._sort(elements)}
                        else:
                            # nested dict
                            new__elmts = {k: self._chemical_element_list_to_dict(v) for k, v in elements.items()}
                            if not self.__validate_distinctness(new__elmts):
                                print("Warning: Chose distinctness, but supplied non-distinct groups. Not stored.")
                                self.__elmts = {"": {}}
                    else:
                        # try converting unknown types of 'elements' into a list
                        try:
                            self.__elmts = {"": self._chemical_element_list_to_dict(list(elements))}
                        except TypeError as err:
                            raise err(f"Argument is a {type(elements)}, but must be a list, dict, tuple or set.")

                # init data: an object store associated with each elmt group
                group_names = self.groups()
                if not group_names:
                    self.__data = {"": None}
                else:
                    self.__data = {group_name: None for group_name in group_names}

    def is_flat(self):
        """True if 'elmts' is dict of chemical elements, False if a dict of groups of such.
        """
        return (not list(self.__elmts.keys())) or (list(self.__elmts.keys()) == [""]) or (len(self.__elmts.keys()) == 1)

    @property
    def elmts(self):
        """If flat, returns list of chemical elements, else groups of such.

        Note that read-only, returns deepcopy.
        """
        import copy
        if self.is_flat():
            key = list(self.__elmts.keys())[0]
            return copy.deepcopy(self.__elmts[key])
        else:
            return copy.deepcopy(self.__elmts)

    @property
    def data(self):
        """Returns full 'data' object storage with all groups.

        Note that if nested, is read-only, returns deepcopy.
        """
        import copy
        if self.is_flat():
            return self.data
            # key = list(self.__elmts.keys())[0]
            # return self.__data[key]
            # return copy.deepcopy(self.__data[key])
        else:
            # return self.__data
            return copy.deepcopy(self.__data)

    def get_data(self, group_name: str):
        """Returns 'data' object storage for given group.
        """
        if self.is_flat():
            return self.data
        else:
            return self.__data[group_name]

    def set_data(self, group_name: str, an_object):
        """Set a data item in the 'data' object storage.

        :return: previously stored data item if present
        """
        if self.is_flat():
            key = list(self.__elmts.keys())[0]
        else:
            if group_name not in self.groups():
                raise KeyError("Supplied group name does not exist.")
            key = group_name
        dump = self.__data.pop(key, None)
        self.__data[key] = an_object
        return dump

    def pop_data(self, group_name: str):
        return self.set_data(group_name, None)

    def __eq__(self, other: object) -> bool:
        """Overload '==' operator and '!=' operator. Check for equality.

        Note: this DOES take grouping into account.
        """
        if isinstance(other, ChemicalElements):
            return self.__elmts == other.__elmts
        return False

    def __le__(self, other):
        """Overload '<=' operator. Is this subset of other.

        True if this flattened is a subset of the other (ie, ignoring grouping).

        :param other: possible superset
        :type other: ChemicalElements
        """
        assert isinstance(other, ChemicalElements)
        return (other + self).flatten(as_dict=False) == other.flatten(as_dict=False)

    def __lt__(self, other):
        """Overload '<' operator. Is this true subset of other.

        True if this flattened is a subset of the other (ie, ignoring grouping).

        :param other: possible superset
        :type other: ChemicalElements
        """
        return (self <= other) and (self.count() < other.count())

    def __ge__(self, other):
        """Overload '>=' operator. Is other subset of this.

        True if other flattened is a subset of this (ie, ignoring grouping).

        :param other: possible subset
        :type other: ChemicalElements
        """
        assert isinstance(other, ChemicalElements)
        return (self + other).flatten(as_dict=False) == self.flatten(as_dict=False)

    def __gt__(self, other):
        """Overload '>' operator. Is other true subset of this.

        True if other flattened is a true subset of this (ie, ignoring grouping).

        :param other: possible subset
        :type other: ChemicalElements
        """
        assert isinstance(other, ChemicalElements)
        return (self >= other) and (self.count() > other.count())

    def __getitem__(self, group_name_or_symbol):
        """Overload '[]' operator, getter.

        If flat, returns symbol's atomic number, if nested, returns group of elements.
        If flat, also allows inverted input: atomic_number, return symbol.
        """
        if isinstance(group_name_or_symbol, str):
            return self.elmts[group_name_or_symbol]
        elif isinstance(group_name_or_symbol, int):
            if self.is_flat():
                return self.invert()[group_name_or_symbol]
            else:
                raise KeyError("Querying [atomic_number] not possible for nested elmts.")
        else:
            raise KeyError(f"Unsupported key/value type '{type(group_name_or_symbol)}'.")

    def __validate_distinctness(self, new__elmts):
        import copy
        if not self.is_flat() and self.distinct:
            elmt_count = 0
            for group_name, group in new__elmts.items():
                elmt_count += len(group.keys())
            distinct_count = len(set().union(*[set(group.keys())
                                               for group_name, group in new__elmts.items()]))
            if elmt_count == distinct_count:
                self.__elmts = copy.deepcopy(new__elmts)
                return True
            else:
                return False
        else:
            self.__elmts = copy.deepcopy(new__elmts)
            return True

    def __setitem__(self, group_name, elements):
        """Overload '[]' operator, setter. Sets group of list of chemical elements.
        
        Note: this overrides the current elmts in that group, or all if elmts is flat.
        For adding elements, use add_elements().

        :param group_name: a name for this list of elements
        :param elements: examples ['He', 'Na']; [2, 11]; {'bla':['Ca'], 'ble;"['Cu, 'Au'].
        :type elements: dict, list, tuple or set
        """
        self.__elmts[group_name] = self._chemical_element_list_to_dict(elements)
        self.__data[group_name] = None

    def items(self):
        """Overload iteration key-value. Note that iterates over groups if not flat."""
        for group_or_symbol in self.elmts:
            yield (group_or_symbol, self.elmts[group_or_symbol])

    def __iter__(self):
        """Overload iteration key. Note that iterates over groups if not flat."""
        return iter(self.elmts)

    def __contains__(self, item):
        """'Overload 'in' operator. Calls count()."""
        _item = item
        if isinstance(_item, (str, int)):
            return self.count(_item)
        elif isinstance(item, (list, dict)):
            _item = ChemicalElements(elements=_item)
        if isinstance(_item, ChemicalElements):
            return _item <= self
        return False

    def __add__(self, other):
        """Overload '+' operator. Addition = set union of two possibly nested dicts."""
        assert isinstance(other, ChemicalElements)
        distinct = self.distinct and other.distinct
        union = ChemicalElements(None, empty=True, distinct=distinct)
        for group_name, group in self.__elmts.items():
            union.add_elements(group, group_name)
        for group_name, group in other.__elmts.items():
            union.add_elements(group, group_name)
        return union

    def __sub__(self, other):
        """Overload '-' operator. Subtraction = set left-difference of possibly nested dicts."""
        assert isinstance(other, ChemicalElements)
        distinct = self.distinct and other.distinct
        difference = ChemicalElements(None, empty=True, distinct=distinct)
        for group_name, group in self.__elmts.items():
            difference.add_elements(group, group_name)
        for group_name, group in other.__elmts.items():
            difference.remove_elements(group, group_name)
        return difference

    def keys(self):
        """If flat, returns symbols, if nested, returns group names, like groups().
        """
        return self.elmts.keys()

    def values(self):
        """If flat, returns atomic numbers, if nested, returns elements of group.
        """
        return self.elmts.values()

    def clear(self, selected_groups: list = None, delete_empty_groups: bool = True):
        """Clear selected groups, optionally delete empty groups.

        Note that objects stored in 'data' get lost.

        :param selected_groups: if None, clear all groups.
        :type selected_groups: list of strings
        :param delete_empty_groups: remove group_name key from elmts if group empty
        """
        if self.is_flat():
            self.__elmts = {"": {}}
            self.__data = {"": {}}
        else:
            if delete_empty_groups:
                for group_name in selected_groups:
                    self.__elmts.pop(group_name, None)
                    self.__data.pop(group_name, None)
                if not self.__elmts.keys():
                    self.__elmts = {"": {}}
                    self.__data = {"": {}}
            else:
                for group_name in selected_groups:
                    if group_name in self.__elmts:
                        self.__elmts[group_name] = {}
                        self.__data[group_name] = None

    def flatten(self, selected_groups: list = None, in_place: bool = False, as_dict=True):
        """Return flattened dict of group selection, or replace elmts with it in-place.
        
        Note: if in_place, all objects stored in 'data' will be lost.
        
        :param selected_groups: If nested and not specified, include all groups.
        :type selected_groups: list of strings
        :param in_place: True: replace elmts with flattened dict.
        :param as_dict: True: return as dict, False: as ChemicalElements
        :return: if not in-place, return flattened dict, else None.
        :rtype: None, dict, or ChemicalElements
        """
        if self.is_flat() and not in_place:
            import copy
            if as_dict:
                return copy.deepcopy(self.elmts)
            else:
                return ChemicalElements(self.elmts)
        elif not self.is_flat():
            data_selection = self.select_groups(selected_groups)
            flattened = {}
            for group_name, group in data_selection.items():
                for sym, num in group.items():
                    flattened[sym] = num
            if in_place:
                self.__elmts = {"": flattened}
                self.__data = {"": None}
            else:
                if as_dict:
                    return flattened
                else:
                    return ChemicalElements(flattened)

    def count_groups(self, selected_groups: list = None):
        """If flat, return element count. If nested, return dict with element count per group.

        :param selected_groups: only is nested: only count specified groups.
        :type selected_groups: list of strings
        """
        if self.is_flat():
            return len(self.elmts.keys())
        else:
            return {k: len(v.keys()) for k, v in self.select_groups(selected_groups).items()}

    def count(self, item=None, group_name: str = None) -> int:
        """Count occurrences of symbol or atomic_number in elmts.

        If nothing specified, count all distinct elements in elmts (sum over all groups).
        If only group_name specified, count all elements in group.
        If only item specified, count element occurrences across all groups.
        If both specified, count item in this group (1 or 0).

        :param item: symbol (case sensitive) or atomic_number
        :type item: str or int
        :param group_name: Optional if elmts is flat, obligatory if nested.
        :type group_name: str
        :return: if not return_dict: 0 if absent, >0 if present
        """
        if item is None:
            if group_name in self.__elmts:
                return len(self.__elmts[group_name].keys())
            elif group_name is None:
                return len(self.flatten().keys())
            else:  # group_name but not in self.__elmts
                raise KeyError(f"Group '{group_name}' is not in elmts.")

        def contains_item(a_dict):
            if isinstance(item, str):
                return int(item in a_dict)
            elif isinstance(item, int):
                return int(item in a_dict.values())
            else:
                return int(False)

        if self.is_flat():
            return contains_item(self.elmts)
        elif group_name is None:  # and nested
            # check all groups
            count = 0
            for group in self.elmts.values():
                count += contains_item(group)
            return count
        else:  # nested and group_name
            return contains_item(self.elmts[group_name])

    def groups(self):
        """Returns list of group names. If elmts is flat (only one group), will return empty list.
        """
        if self.is_flat():
            return []
        else:
            return list(self.elmts.keys())

    def select_groups(self, selected_groups: list = None, include_special_elements: bool = True, as_dict=True):
        """Return only selected groups from elmts. Always returns a copy.

        :param selected_groups: If flat and not specified, return the empty name group of flat elmts.
        :type selected_groups: list of strings
        :param include_special_elements: False: remove special elements if any defined. This returns a copy always.
        :param as_dict: True: return as dict, False: as ChemicalElements
        :return: subset of elmts
        :rtype: dict
        """
        import copy

        elements = copy.deepcopy(self.__elmts)

        if not include_special_elements:
            for group_name, group in elements.items():
                for special_symbol in self._special_elements:
                    elements[group_name].pop(special_symbol, None)

        if not selected_groups or (set(self.groups()) == set(selected_groups)):
            # always return non-flat view, ie if flat, dict with one group
            if as_dict:
                return elements
            else:
                return self
        else:
            a_dict = {data_group: elmts for data_group, elmts in elements.items()
                      for selection_group in selected_groups if data_group == selection_group}
            if as_dict:
                return a_dict
            else:
                elmts = ChemicalElements(empty=True)
                for group_name, group in a_dict.items():
                    elmts.add_elements(elements=group, group_name=group_name)
                return elmts

    def rename_group(self, old_group_name, new_group_name):
        """Rename a group.
        """
        group = self.__elmts.pop(old_group_name)
        self.add_elements(group, new_group_name)

        data = self.__data.pop(old_group_name)
        self.__data[new_group_name] = data

    def group_difference(self, group_name1: str, group_name2: str) -> dict:
        """Return difference between two groups.

        If one is the true superset, the set left difference is returned (what is in the one that is not in the other).

        If not, the symmetrical difference is returned.
        """
        if self.is_flat():
            raise KeyError("Elmts is flat, no groups to compare.")
        if not ((group_name1 in self.groups()) and (group_name2 in self.groups())):
            raise KeyError(f"Elmts has no groups '{group_name1}', '{group_name2}'.")
        distinct = self.distinct
        elmts_group1 = ChemicalElements(self[group_name1], distinct=distinct)
        elmts_group2 = ChemicalElements(self[group_name2], distinct=distinct)
        if elmts_group1 == elmts_group2:
            print(f"'{group_name1}' equal to '{group_name2}'")
            return ChemicalElements(empty=True, distinct=distinct)
        elif elmts_group1 > elmts_group2:
            print(f"'{group_name1}' true superset of '{group_name2}'")
            return elmts_group1 - elmts_group2
        elif elmts_group1 < elmts_group2:
            print(f"'{group_name1}' true subset of '{group_name2}'")
            return elmts_group2 - elmts_group1
        else:
            print(f"'{group_name1}', '{group_name2}' not subsets of each other, return symmetrical difference")
            return elmts_group1.symmetrical_difference(elmts_group2)

    def compare_groups(self, group_name1, group_name2):
        from deepdiff import DeepDiff
        if self.is_flat():
            raise KeyError("Elmts is flat, no groups to compare.")
        if not ((group_name1 in self.groups()) and (group_name2 in self.groups())):
            raise KeyError(f"Elmts has no groups '{group_name1}', '{group_name2}'.")
        return DeepDiff(self[group_name1], self[group_name2], ignore_order=True, ignore_numeric_type_changes=True)

    def invert(self, group_name: str = None):
        """If flat, return inverted element dict symbol : atomic number to num:sym, or group elmt dict if nested.

        :return: inverted chemical elements dict
        :rtype: dict
        """
        if self.is_flat():
            elmts = self.elmts if self.is_flat() else self.__elmts[group_name]
        else:
            if group_name:
                elmts = self.__elmts[group_name]
            else:
                raise KeyError("Is nested, but did not supply group name.")
        return {v: k for k, v in elmts.items()}

    def get_groups_for(self, symbol: str):
        """Find the group names for the groups the chemical element is contained in.
        :param symbol: chemical element symbol
        :return: list of group_names or None if no find or if flat
        """
        if self.is_flat():
            return []
        else:
            group_names = [group_name for group_name in self.elmts if symbol in self.elmts[group_name]]
            if group_names:
                return group_names
            else:
                return []

    def add_elements(self, elements, group_name: str = ""):
        """Add elements. Elmts get resorted afterwards.

        :param elements: list of symbols str or atomic numbers int
        :type elements: list or dict
        :param group_name: if this nested, add to this' group. ignored if this flat.
        :type group_name: str
        """
        if isinstance(elements, dict):
            # just assume it's a element sym:num dict
            elements, _ = self._validate(elements)
            elmts = elements
        else:
            elmts = self._chemical_element_list_to_dict(elements, sort=False)

        import copy
        new__elmts = copy.deepcopy(self.__elmts)
        if group_name not in self.__elmts:
            if not self.is_flat() and not group_name:
                # generate a random group name and add elements
                from masci_tools.util.python_util import random_string
                group_name = "UNNAMED_" + random_string(5)
                print(f"Warning: adding nested and flat ChemicalElements, "
                      f"adding flat elements to new group '{group_name}'")
            new__elmts[group_name] = {}
        for sym, num in elmts.items():
            new__elmts[group_name][sym] = num
        if not self.__validate_distinctness(new__elmts):
            print("Warning: Chose distinctness, added elements would violate. Not added.")
        else:
            self.__elmts[group_name] = self._sort(self.__elmts[group_name])
            if not group_name in self.__data:
                self.__data[group_name] = None

    def expand_allowed_elements(self, symbol: str, atomic_number: int):
        """Expand allowed elements definition by a 'special element' which is not in the standard periodic table.

        Example: In some applications, 'X':0 element is used to represent free space.

        Note: This will not add the element to elmts, but enable the addition or removal of the special element.
        """
        assert isinstance(symbol, str) and isinstance(atomic_number, int)

        info_msg_prefix = f"INFO: Requested to expand allowed elements definition by special element " \
                          f"{{{symbol} : {atomic_number}}}. "

        def _remove_special_element_from_definition(symbol, atomic_number):
            self._special_elements.pop(symbol)
            self._special_elements_inv.pop(atomic_number)
            self._pte.pop(symbol)
            self._pte_inv.pop(atomic_number)

        # check symbol
        if symbol in self._pte:
            if symbol not in self._special_elements:
                print(info_msg_prefix +
                      f"Symbol is a standard element of the periodic table. "
                      f"I will not expand definition by this element.")
                return
            else:
                # now need to check if the stored special element with the same symbol has a different atomic number
                # if so, remove it
                stored_atomic_number = self._special_elements[symbol]
                if atomic_number != stored_atomic_number:
                    print(info_msg_prefix +
                          f"Found stored special element '{stored_atomic_number}' with same symbol. "
                          f"I will replace the latter with the former.")
                    _remove_special_element_from_definition(symbol=symbol, atomic_number=stored_atomic_number)

        # check atomic number
        if atomic_number in self._pte_inv:
            if atomic_number not in self._special_elements_inv:
                print(info_msg_prefix +
                      f"Atomic number is that of a standard element of the periodic table. "
                      f"I will not expand definition by this element.")
                return
            else:
                # now need to check if the stored special element with the same atomic number has a different symbol
                # if so, remove it
                stored_symbol = self._special_elements_inv[atomic_number]
                if symbol != stored_symbol:
                    print(info_msg_prefix +
                          f"Found stored special element '{stored_symbol}' with same atomic number. "
                          f"I will replace the latter with the former.")
                    _remove_special_element_from_definition(symbol=stored_symbol, atomic_number=atomic_number)

        # okay, now finally clear to expand allowed element definition
        self._special_elements[symbol] = atomic_number
        self._special_elements_inv[atomic_number] = symbol
        self._pte[symbol] = atomic_number
        self._pte_inv[atomic_number] = symbol
        self._sort(self._pte)
        self._sort(self._pte_inv, by_key=True)

    def remove_elements(self, elements, group_name: str = None, delete_empty_groups: bool = True):
        """Remove elements. Elmts get resorted afterwards.

        If a now empty group is deleted and corresponding 'data' entry held an object, that object
        is returned.

        :param elements: list of symbols str or atomic numbers int
        :type elements: list
        :param group_name: if flat, ignored, if nested, remove from this group
        :type group_name: str
        :param delete_empty_groups: remove group_name key from elmts if group empty
        """
        if isinstance(elements, dict):
            # just assume it's a element sym:num dict
            elements, _ = self._validate(elements)
            elmts = elements
        else:
            elmts = self._chemical_element_list_to_dict(elements, sort=False)
        if group_name is None and self.is_flat():
            group_name = list(self.__elmts.keys())[0]
        if group_name in self.__elmts:
            for sym in elmts:
                self.__elmts[group_name].pop(sym, None)
            if self.__elmts[group_name]:
                self.__elmts[group_name] = self._sort(self.__elmts[group_name])
            if not self.__elmts[group_name] and delete_empty_groups:
                self.__elmts.pop(group_name, None)
                self.__data.pop(group_name, None)
            if not self.__elmts.keys():
                self.__elmts = {"": {}}
                self.__data = {"": {}}
        else:
            print(f"Warning: no group '{group_name}' present in elmts. Nothing removed.")

    def union(self, other):
        return self.__add__(other)

    def difference(self, other):
        return self.__sub__(other)

    def complement(self, selected_groups: list = None):
        """Return set complement of selected groups of elmts with respect to the set of all possible elements.

        Computes union of elements from selected groups, then complement thereof.

        :param selected_groups: If flat and not specified, use all elements in flat elmts.
        :type selected_groups: list of strings
        :return: complement, flat not nested
        :rtype: ChemicalElements
        """
        data_selection = self.select_groups(selected_groups)
        all_elmts = [sym for elmts in data_selection.values() for sym in elmts]
        complement = list(set(self._pte.keys()) - set(all_elmts))
        complement = self._chemical_element_list_to_dict(complement)
        if not complement:
            # neeed this since supplying 'nothing' to constructor fills whole table by default
            return ChemicalElements(empty=True, distinct=self.distinct)
        else:
            return ChemicalElements(list(complement.keys()), distinct=self.distinct)

    def intersection(self, other):
        """Create intersection of two ChemicalElements objects as a new one.

        :param other: other elements container
        :type other: ChemicalElements
        :return: new elements container
        :rtype: ChemicalElements
        """
        assert isinstance(other, ChemicalElements)
        distinct = self.distinct and other.distinct
        intersection = list(set(self.flatten().keys()) & set(other.flatten().keys()))
        return ChemicalElements(intersection, distinct=distinct)

    def symmetrical_difference(self, other):
        """Create intersection of two ChemicalElements objects as a new one.

        :param other: other elements container
        :type other: ChemicalElements
        :return: new elements container
        :rtype: ChemicalElements
        """
        assert isinstance(other, ChemicalElements)
        return (self - other).union(other - self)

    def __repr__(self):
        """For console output."""
        return self.to_string()

    def to_string(self, selected_groups: list = None, indent=4):
        """Print elmts dict nicely indented with linebreaks.
        """
        import json
        elmts = self.select_groups(selected_groups)
        if len(elmts.keys()) == 1:
            key = list(elmts.keys())[0]
            elmts = elmts[key]
        return json.dumps(elmts, indent=indent)

    def print(self, selected_groups: list = None, indent=4):
        """Print elmts dict nicely indented with linebreaks.
        """
        print(self.to_string(selected_groups, indent))

    def to_file(self, filepath):
        """Save elements to file. 'data' is ignored.

        :param filepath: filepath
        :type filepath: pathlib.Path
        """
        import json
        with open(filepath, 'w') as file:
            file.write(json.dumps(self.__elmts))

    def plot(self, selected_groups: list = [], selection_name: str = "",
             unselected_name: str = "Unspecified", output="notebook"):
        """Plot a periodic table, elements optionally grouped by group colors.
        Element groups can either be None, that gives the normal periodic table. Or it can be a list of element symbols
        ('H', 'Ca' and so on), that will print the periodic table with those elements highlighted and the rest greyed
        out. Or it can be a dict of element groups, where the key serves as group name. Then each group gets its own
        color in the periodic table, and a color legend is printed.

        :param selected_groups: If flat and not specified, use all elements in flat elmts, else if nested, a group subset.
        :type selected_groups: list of strings
        :param selection_name: name for the whole element selection
        :param unselected_name: legend name for group of unselected, greyed out elements
        :param output: bokeh plotting output. supported: 'notebook'.
        """
        from matplotlib import colors
        import seaborn as sns
        import bokeh.plotting
        import mendeleev.plotting

        if output == "notebook":
            bokeh.plotting.output_notebook()
        else:
            raise NotImplementedError(f"Output for {output} not implemented.")

        pte = mendeleev.get_table('elements')
        data_selection = self.select_groups(selected_groups, include_special_elements=False)

        # first make another group with all ungrouped elements, to draw them greyed out
        # note: can't use complement() for this, in case only subset of elmts selected
        all_elements = [el.symbol for el in mendeleev.elements.get_all_elements()]
        selected_elements = [element for group in list(data_selection.values()) for element in group]
        unselected_elements = self._chemical_element_list_to_dict(set(all_elements) - set(selected_elements))
        data_selection[unselected_name] = unselected_elements

        # now make a new attribute(column) and label each element according to group
        for group_name, group in data_selection.items():
            for symbol in group.keys():
                pte.loc[pte['symbol'] == symbol, [selection_name]] = group_name
        # now make another attribute and map the newly created attribute to a respective color
        group_names = list(data_selection.keys())
        cmap = {g: colors.rgb2hex(c) for g, c in zip(group_names, sns.color_palette('deep'))}
        grey = '#bfbfbf'
        cmap[unselected_name] = grey

        # if there is more than one group, plot a legend
        if len(data_selection.keys()) > 2:
            masci_tools.vis.plot_methods.plot_colortable(cmap, selection_name)

        colorby_attribute = f"{selection_name}_color"
        pte[colorby_attribute] = pte[selection_name].map(cmap)
        # and plot
        mendeleev.plotting.periodic_plot(pte, colorby=colorby_attribute, title=selection_name)

    def _sort(self, a_dict, by_key=False):
        """Sorts chemical element dict by atomic number.

        :param a_dict: dict key = symbol str : value = atomic_number int
        :type a_dict: dict
        :param by_key: sort by key or value
        :type by_key: bool
        :return: sorted dict
        :rtype: dict
        """
        for_k0_for_v1 = not int(by_key)
        return {k: v for k, v in sorted(a_dict.items(), key=lambda item: item[for_k0_for_v1])}

    def _validate(self, elements):
        """Checks that no non-valid elements are present.
        """
        list_types = (list, tuple, set)
        if isinstance(elements, list_types):
            key_is_symbol = None
            if all([isinstance(sym, str) for sym in elements]):
                assert all([elem in self._pte for elem in elements])
                key_is_symbol = True
            elif all([isinstance(num, int) for num in elements]):
                assert all([elem in self._pte_inv for elem in elements])
                key_is_symbol = False
            else:
                raise TypeError("received a list, but need list of all atom names or numbers")
            return elements, key_is_symbol
        elif isinstance(elements, dict):
            # just assume it is a flat dict
            # i know this is not the most efficient way to test this

            # first, invert dict if needed to simplify cases
            if any([isinstance(k, int) for k in elements.keys()]):
                assert all([isinstance(k, int) for k in elements.keys()])
                elements = {v: k for k, v in elements.items()}

            # now assume dict is sym:num, else raise error
            if any([isinstance(k, str) for k in elements.keys()]):
                assert all([isinstance(k, str) for k in elements.keys()])
                assert all([isinstance(v, int) for v in elements.values()])
                key_is_symbol = True
                return elements, key_is_symbol
            else:
                raise TypeError("received a dict, but need chem.elmt. dict sym:num or num:sym")
        else:
            raise TypeError("Elements list is not any of list,tuple,set,dict.")

    def _chemical_element_list_to_dict(self, elements, sort=True):
        """Converts list/set/tuple of chemical elements into dict symbol : atomic_number.
        :type elements: list, tuple or set of 1) symbols str or 2) atomic numbers int
        :return: dict sym:num
        :rtype: dict
        """
        elements, key_is_sym = self._validate(elements)
        if key_is_sym:
            a_dict = {sym: num for (sym, num) in self._pte.items() if sym in elements}
        else:
            a_dict = {sym: num for (sym, num) in self._pte.items() if num in elements}

        if sort:
            return self._sort(a_dict)
        else:
            return a_dict


@dc.dataclass
class PeriodicTable:
    from masci_tools.util import python_util

    table: ChemicalElements = python_util.dataclass_default_field(ChemicalElements())
    crystal: ChemicalElements = python_util.dataclass_default_field(ChemicalElements({
        'fcc': ['Ir', 'Pd', 'Pb', 'Pt', 'Al', 'Cu', 'Ca', 'Ag', 'Au', 'Sr', 'Mn', 'Ni'],
        'bcc': ['Ba', 'Cr', 'Cs', 'Fe', 'K', 'Mo', 'Nb', 'Rb', 'Ta', 'V', 'W'],
        'hcp': ['Be', 'Cd', 'Co', 'He', 'Hf', 'Mg', 'Os', 'Re', 'Ru', 'Sc', 'Tc', 'Ti', 'Tl', 'Y', 'Zn', 'Zr'],
        'diamond': ['Ge', 'Si', 'Sn'],
        'rhombohedral': ['As', 'Bi', 'Sb']
    }))
    magnet: ChemicalElements = python_util.dataclass_default_field(ChemicalElements({
        'ferromagnetic': ['Fe', 'Co', 'Ni'],
        'antiferromagnetic': ['Cr', 'Mn', 'O']
    }))



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains a generic HDF5 reader
"""
import io
import h5py
from collections import namedtuple
import warnings
import logging

Transformation = namedtuple('Transformation', ['name', 'args', 'kwargs'])
AttribTransformation = namedtuple('AttribTransformation', ['name', 'attrib_name', 'args', 'kwargs'])

logger = logging.getLogger(__name__)


class HDF5Reader:
    """Class for reading in data from hdf5 files using a specified recipe

    :param file: filepath to hdf file or opened file handle (mode 'rb')
    :param move_to_memory: bool if True after reading and transforming the data
                           all leftover h5py.Datasets are moved into np.arrays

    The recipe is passed to the :py:meth:`HDF5Reader.read()` method and consists
    of a dict specifiying which attributes and datasets to read in and how to transform them

    Each attribute/dataset entry corresponds to one entry point in the given `.hdf` file
    Available transformations can either be found in :py:mod:`~masci_tools.io.parsers.hdf5.transforms`
    or can be defined by the user with the :py:func:`~masci_tools.io.parsers.hdf5.transforms.hdf5_transformation`
    decorator

    Basic Usage:

    .. code-block:: python

        from masci_tools.io.parsers.hdf5 import HDF5Reader
        import masci_tools.io.parsers.hdf5.recipes as recipes

        #This example shows the usage for producing data from a bandstructure calculation
        #in Fleur
        with HDF5Reader('/path/to/hdf/banddos.hdf') as h5reader:
            data, attributes = h5reader.read(recipe=recipes.FleurBands)
        print(data, attributes)

    """

    def __init__(self, file, move_to_memory=True):

        self._file = file

        if isinstance(self._file, io.IOBase):
            self._filename = self._file.name
        else:
            self._filename = self._file

        if not self._filename.endswith('.hdf'):
            logger.exception('Wrong File Type for %s: Got %s', self.__class__.__name__, self._filename)
            raise ValueError(f'Wrong File Type for {self.__class__.__name__}: Got {self._filename}')

        logger.info('Instantiated %s with file %s', self.__class__.__name__, self._filename)

        self._move_to_memory = move_to_memory
        self._h5_file = None

    def __enter__(self):
        self._h5_file = h5py.File(self._file, 'r')
        logger.debug('Opened h5py.File with id %s', self._h5_file.id)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._h5_file.close()
        logger.debug('Closed h5py.File with id %s', self._h5_file.id)

    def _read_dataset(self, h5path, strict=True):
        """Return in the dataset specified by the given h5path

        :param h5path : str, HDF5 group path in file.
        :param strict : bool if no dataset is at the path and strcit is True raise Error
                        else return None

        :returns: h5py.Dataset or None

        :raises: ValueError if no dataset is at path and strict is True
        """
        if h5path in ('/', ''):
            if strict:
                return None
            else:
                pass

        logger.debug('Reading dataset from path %s', h5path)

        dset = self._h5_file.get(h5path)
        if dset is not None:
            return dset
        elif strict:
            logger.exception('HDF5 input file %s has no Dataset at %s.', self._file, h5path)
            raise ValueError(f'HDF5 input file {self._file} has no Dataset at {h5path}.')
        return None

    def _transform_dataset(self, transforms, dataset, attributes=None, dataset_name=None):
        """
        Transforms the given dataset with the given list of tasks

        :param transforms: list of namedtuples defining the tasks to perform
        :param dataset: h5py.Dataset, on which to perform the operations
        :param attributes: dict of previously processed attributes.
                           Only available to the entries in the datasets
                           section of the recipe. This allows for operations with
                           the previously parsed attributes

        :returns: the dataset with all the transformations applied
        """
        transformed_dset = dataset
        for spec in transforms:

            args = spec.args
            if spec.name in self._attribute_transforms:
                if attributes is None:
                    raise ValueError('Attribute transform not allowed for attributes')
                attrib_value = attributes[spec.attrib_name]
                args = attrib_value, *args

            logger.debug('Applying transformation %s to dataset %s of type %s', spec.name, dataset_name,
                         type(transformed_dset))

            try:
                transformed_dset = self._transforms[spec.name](transformed_dset, *args, **spec.kwargs)
            except Exception as err:
                logger.exception(str(err))
                raise

        return transformed_dset

    @staticmethod
    def _unpack_dataset(output_dict, dataset_name):
        """
        Unpack the entires of the dictionary dataset in the entry dataset_name into the
        output_dict

        :param output_dict: dict with the dataset entries
        :param dataset_name: key of the dataset to unpack into output_dict

        :returns: output_dict with the entries of dataset_name unpacked
        """

        logger.debug('Unpacking dict dataset %s after transformations', dataset_name)

        if not isinstance(output_dict[dataset_name], dict):
            raise ValueError(f'{dataset_name} cannot be unpacked: Got {type(output_dict[dataset_name])}')

        unpack_dict = output_dict.pop(dataset_name)

        if unpack_dict.keys() & output_dict.keys():
            raise ValueError('Unpacking would result in lost information: \n'
                             f"Intersection of keys: '{unpack_dict.keys().intersection(output_dict.keys())}'")

        return {**output_dict, **unpack_dict}

    def read(self, recipe=None):
        """Extracts datasets from HDF5 file, transforms them and puts all into a namedtuple.

        :param recipe: dict with the format given in :py:mod:`~masci_tools.io.parsers.hdf5.recipes`

        :returns: two dicts with the datasets/attributes read in and transformed according to the recipe
        """
        from itertools import chain
        from masci_tools.io.hdf5_util import read_hdf_simple

        logger.info('Started reading HDF file: %s', self._filename)

        if recipe is None:
            msg = 'Using the HDF5Reader without a recipe falling back to simple HDF reader'
            logging.warning(msg)
            warnings.warn(msg)
            res = read_hdf_simple(self._file)
            logger.info('Finished reading .hdf file')
            return res

        datasets = recipe.get('datasets', {})
        attributes = recipe.get('attributes', {})

        # remove entries whose key is an empty string
        h5paths = {item['h5path'] for item in chain(datasets.values(), attributes.values())}
        extracted_datasets = {h5path: self._read_dataset(h5path) for h5path in h5paths}

        output_attrs = {}
        for key, val in attributes.items():
            transforms = val.get('transforms', [])
            output_attrs[key] = self._transform_dataset(transforms, extracted_datasets[val['h5path']], dataset_name=key)
            if val.get('unpack_dict', False):
                try:
                    output_attrs = self._unpack_dataset(output_attrs, dataset_name=key)
                except Exception as err:
                    logger.exception(str(err))
                    raise

        output_data = {}
        for key, val in datasets.items():
            transforms = val.get('transforms', [])
            output_data[key] = self._transform_dataset(transforms,
                                                       extracted_datasets[val['h5path']],
                                                       attributes=output_attrs,
                                                       dataset_name=key)
            if val.get('unpack_dict', False):
                try:
                    output_data = self._unpack_dataset(output_data, dataset_name=key)
                except Exception as err:
                    logger.exception(str(err))
                    raise

        if self._move_to_memory:
            logger.debug('Moving remaining h5py.Datasets to memory')
            try:
                self._transforms['move_to_memory'](output_data)
                self._transforms['move_to_memory'](output_attrs)
            except Exception as err:
                logger.exception(str(err))
                raise

        logger.info('Finished reading HDF file: %s', self._filename)

        return output_data, output_attrs



# -*- coding: utf-8 -*-
"""
Test of the consistency the input schema dictionaries with the SchemaFiles in the same folder
"""
import pytest
import os

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
SCHEMA_DIR = '../io/parsers/fleur/fleur_schema'

#Collect all schemas from the folder
schema_versions = {'inp': [], 'out': []}
for root, dirs, files in os.walk(os.path.abspath(os.path.join(CURRENT_DIR, SCHEMA_DIR))):
    for folder in dirs:
        if '0.' in folder:
            schema_versions['inp'].append(folder)
            if int(folder.split('.')[1]) >= 33 or folder in ('0.31', '0.30', '0.29'):
                schema_versions['out'].append(folder)


def test_inpschema_dict_structure():
    """
    Test the types of the keys in the inpschemadict
    """
    from masci_tools.io.parsers.fleur.fleur_schema import InputSchemaDict
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict
    from masci_tools.util.lockable_containers import LockableDict, LockableList

    inputschema = InputSchemaDict.fromVersion('0.34')

    EXPECTED_TYPES = {
        'tag_paths': CaseInsensitiveDict,
        'unique_attribs': CaseInsensitiveDict,
        'unique_path_attribs': CaseInsensitiveDict,
        'other_attribs': CaseInsensitiveDict,
        'attrib_types': CaseInsensitiveDict,
        'simple_elements': CaseInsensitiveDict,
        '_basic_types': LockableDict,
        'root_tag': str,
        'inp_version': str,
        'omitt_contained_tags': LockableList,
        'tag_info': LockableDict
    }

    for key, val in inputschema.items():
        assert isinstance(val, EXPECTED_TYPES[key])
        if EXPECTED_TYPES[key] != str:
            assert val._locked
            if EXPECTED_TYPES[key] != CaseInsensitiveDict:
                assert not isinstance(val,
                                      CaseInsensitiveDict)  #since CaseInsensitiveDict is a subclass of LockableDict


def test_outschema_dict_structure():
    """
    Test the types of the keys in the inpschemadict
    """
    from masci_tools.io.parsers.fleur.fleur_schema import OutputSchemaDict
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict
    from masci_tools.util.lockable_containers import LockableDict, LockableList

    outputschema = OutputSchemaDict.fromVersion('0.34')

    EXPECTED_TYPES = {
        'tag_paths': CaseInsensitiveDict,
        'unique_attribs': CaseInsensitiveDict,
        'unique_path_attribs': CaseInsensitiveDict,
        'other_attribs': CaseInsensitiveDict,
        'iteration_tag_paths': CaseInsensitiveDict,
        'iteration_unique_attribs': CaseInsensitiveDict,
        'iteration_unique_path_attribs': CaseInsensitiveDict,
        'iteration_other_attribs': CaseInsensitiveDict,
        'attrib_types': CaseInsensitiveDict,
        'simple_elements': CaseInsensitiveDict,
        '_basic_types': LockableDict,
        '_input_basic_types': LockableDict,
        'root_tag': str,
        'input_tag': str,
        'out_version': str,
        'inp_version': str,
        'omitt_contained_tags': LockableList,
        'tag_info': LockableDict,
        'iteration_tag_info': LockableDict
    }

    for key, val in outputschema.items():
        assert isinstance(val, EXPECTED_TYPES[key])
        if EXPECTED_TYPES[key] != str:
            assert val._locked
            if EXPECTED_TYPES[key] != CaseInsensitiveDict:
                assert not isinstance(val,
                                      CaseInsensitiveDict)  #since CaseInsensitiveDict is a subclass of LockableDict


@pytest.mark.parametrize('schema_version', schema_versions['inp'])
def test_inpschema_dict(data_regression, schema_version):
    """
    Test the produced inputschema dicts
    """
    from masci_tools.io.parsers.fleur.fleur_schema import InputSchemaDict

    inputschema = InputSchemaDict.fromVersion(version=schema_version)

    data_regression.check(clean_for_reg_dump(inputschema.get_unlocked()))


@pytest.mark.parametrize('inp_version', schema_versions['inp'])
@pytest.mark.parametrize('out_version', schema_versions['out'])
def test_outschema_dict(data_regression, inp_version, out_version):
    """
    Test the fleur_schema_parser_functions to make sure that they match the stored inputschema_dict
    """
    from masci_tools.io.parsers.fleur.fleur_schema import OutputSchemaDict

    outputschema = OutputSchemaDict.fromVersion(version=out_version, inp_version=inp_version)

    data_regression.check(clean_for_reg_dump(outputschema.get_unlocked()))


def clean_for_reg_dump(dict_to_clean):
    """
    Clean for data regression converts CaseInsensitiveFrozenSet to set
    Lockable containers are to be converted berfore via `get_unlocked()`
    """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    for key, val in dict_to_clean.items():
        if isinstance(val, CaseInsensitiveFrozenSet):
            dict_to_clean[key] = set(val)
        elif isinstance(val, dict):
            dict_to_clean[key] = clean_for_reg_dump(val)

    return dict_to_clean



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for kkr-specific plotting functions
"""
import pytest

# prevent issue with not having a display on travis-ci
# this needs to go *before* pyplot imports
import matplotlib
matplotlib.use('Agg')
from matplotlib.pyplot import gcf, title
from masci_tools.io.kkr_read_shapefun_info import read_shapefun
from masci_tools.vis.kkr_plot_shapefun import plot_shapefun
from masci_tools.vis.kkr_plot_dos import dosplot
from masci_tools.vis.kkr_plot_bandstruc_qdos import dispersionplot
from masci_tools.vis.kkr_plot_FS_qdos import FSqdos2D


class Test_kkr_plotting(object):
    """
    Test for KKR plotting functions
    """

    @pytest.mark.mpl_image_compare(baseline_dir='files/voronoi/', filename='test.png')
    def test_plot_shapefun(self):
        # clear previous figure, if still there
        gcf().clear()
        pos, out = read_shapefun('files/voronoi/')
        plot_shapefun(pos, out, 'all')
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_dos_output/', filename='test.png', tolerance=5)
    def test_plot_dos(self):
        gcf().clear()
        dosplot('files/kkr/kkr_run_dos_output/')
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_dos_output/', filename='test2.png')
    def test_plot_dos2(self):
        gcf().clear()
        dosplot('files/kkr/kkr_run_dos_output/', units='eV_rel', nofig=True, allatoms=True, totonly=False)
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_dos_output/', filename='test3.png')
    def test_plot_dos3(self):
        gcf().clear()
        dosplot('files/kkr/kkr_run_dos_output/',
                units='eV_rel',
                nofig=True,
                allatoms=True,
                filled=True,
                normalized=True,
                xyswitch=True,
                color='r')
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_dos_output/', filename='test4.png')
    def test_plot_dos4(self):
        gcf().clear()
        dosplot('files/kkr/kkr_run_dos_output/', units='eV_rel', nofig=True, allatoms=True, lm=list(range(1, 5)))
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_qdos/', filename='test.png')
    def test_plot_qdos(self):
        gcf().clear()
        dispersionplot('files/kkr/kkr_run_qdos', reload_data=True)
        title('')
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_qdos/', filename='test2.png')
    def test_plot_qdos2(self):
        gcf().clear()
        dispersionplot('files/kkr/kkr_run_qdos',
                       reload_data=True,
                       ratios=False,
                       units='eV_rel',
                       clrbar=False,
                       shading='nearest')
        title('')
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/kkr/kkr_run_qdos_FS/', filename='test.png')
    def test_plot_qdos_FS(self):
        gcf().clear()
        FSqdos2D('files/kkr/kkr_run_qdos_FS/', reload_data=True)
        return gcf()



# -*- coding: utf-8 -*-
"""
Test for the case insensitive, immutable set implemented in util
"""
import pytest

TEST_INIT = ['ABC', 'abC', 'RanDomEntry5', 'snake_case_entry', 'camelCaseEntry', 'camelcaseentry']
TEST_MEMBER = [('randomentry5', True, 'RanDomEntry5'), ('snakecaseentry', False, None),
               ('camelcaseentry', True, 'camelCaseEntry'), ('aBc', True, 'ABC'), ('not_existent', False, None)]


@pytest.mark.parametrize('key,is_member,original_case', TEST_MEMBER)
def test_case_insensitive_frozenset_member(key, is_member, original_case):
    """
   Test CaseInsensitiveFrozenSet for correct membership tests and the right stored original case
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    s = CaseInsensitiveFrozenSet(TEST_INIT)

    if is_member:
        assert key in s
        assert s.original_case[key] == original_case
    else:
        assert key not in s


TEST_SET = [{'ABC', 'abC', 'RanDomEntry5', 'snake_case_entry', 'camelCaseEntry', 'camelcaseentry'},
            {'ABC', 'not_existent', 'snakecaseentry'}]

TEST_EQUAL = [True, False]


@pytest.mark.parametrize('test_set,is_equal', zip(TEST_SET, TEST_EQUAL))
def test_case_insensitive_frozenset_eq(test_set, is_equal):
    """
   Test CaseInsensitiveFrozenSet for correct equality
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    s = CaseInsensitiveFrozenSet(TEST_INIT)

    if is_equal:
        assert s == test_set
    else:
        assert s != test_set


TEST_DIFFERENCES = [set(), {'camelcaseentry', 'randomentry5', 'snake_case_entry'}]
TEST_DIFF_CASES = [{}, {
    'randomentry5': 'RanDomEntry5',
    'camelcaseentry': 'camelCaseEntry',
    'snake_case_entry': 'snake_case_entry'
}]


@pytest.mark.parametrize('test_set,exp_diff,exp_case', zip(TEST_SET, TEST_DIFFERENCES, TEST_DIFF_CASES))
def test_case_insensitive_frozenset_difference(test_set, exp_diff, exp_case):
    """
   Test CaseInsensitiveFrozenSet for correct difference
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    s = CaseInsensitiveFrozenSet(TEST_INIT)

    actual = s.difference(test_set)

    assert isinstance(actual, CaseInsensitiveFrozenSet)
    assert s - test_set == actual
    assert actual.original_case == exp_case


TEST_UNIONS = [{'camelcaseentry', 'snake_case_entry', 'randomentry5', 'abc'},
               {'camelcaseentry', 'snake_case_entry', 'randomentry5', 'abc', 'not_existent', 'snakecaseentry'}]
TEST_UNION_CASES = [{
    'camelcaseentry': 'camelCaseEntry',
    'abc': 'ABC',
    'randomentry5': 'RanDomEntry5',
    'snake_case_entry': 'snake_case_entry'
}, {
    'not_existent': 'not_existent',
    'camelcaseentry': 'camelCaseEntry',
    'snakecaseentry': 'snakecaseentry',
    'snake_case_entry': 'snake_case_entry',
    'abc': 'ABC',
    'randomentry5': 'RanDomEntry5'
}]


@pytest.mark.parametrize('test_set,exp_union,exp_case', zip(TEST_SET, TEST_UNIONS, TEST_UNION_CASES))
def test_case_insensitive_frozenset_union(test_set, exp_union, exp_case):
    """
   Test CaseInsensitiveFrozenSet for correct union
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    s = CaseInsensitiveFrozenSet(TEST_INIT)

    actual = s.union(test_set)

    assert isinstance(actual, CaseInsensitiveFrozenSet)
    assert actual == exp_union
    assert s | test_set == actual
    assert actual.original_case == exp_case


TEST_SYM_DIFFERENCES = [set(), {'camelcaseentry', 'randomentry5', 'snake_case_entry', 'not_existent', 'snakecaseentry'}]
TEST_SYMDIFF_CASES = [{}, {
    'snakecaseentry': 'snakecaseentry',
    'randomentry5': 'RanDomEntry5',
    'camelcaseentry': 'camelCaseEntry',
    'not_existent': 'not_existent',
    'snake_case_entry': 'snake_case_entry'
}]


@pytest.mark.parametrize('test_set,exp_diff,exp_case', zip(TEST_SET, TEST_SYM_DIFFERENCES, TEST_SYMDIFF_CASES))
def test_case_insensitive_frozenset_sym_difference(test_set, exp_diff, exp_case):
    """
   Test CaseInsensitiveFrozenSet for correct symmetric difference
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    s = CaseInsensitiveFrozenSet(TEST_INIT)

    actual = s.symmetric_difference(test_set)
    print(actual.original_case)
    assert isinstance(actual, CaseInsensitiveFrozenSet)
    assert actual == exp_diff
    assert s ^ test_set == actual
    assert actual.original_case == exp_case


TEST_INTERSECTIONS = [{'camelcaseentry', 'snake_case_entry', 'randomentry5', 'abc'}, {'abc'}]
TEST_INTERSECTIONS_CASES = [{
    'abc': 'ABC',
    'randomentry5': 'RanDomEntry5',
    'snake_case_entry': 'snake_case_entry',
    'camelcaseentry': 'camelCaseEntry'
}, {
    'abc': 'ABC'
}]


@pytest.mark.parametrize('test_set,exp_intersect,exp_case', zip(TEST_SET, TEST_INTERSECTIONS, TEST_INTERSECTIONS_CASES))
def test_case_insensitive_frozenset_intersection(test_set, exp_intersect, exp_case):
    """
   Test CaseInsensitiveFrozenSet for correct intersection
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet

    s = CaseInsensitiveFrozenSet(TEST_INIT)

    actual = s.intersection(test_set)
    print(actual.original_case)
    assert isinstance(actual, CaseInsensitiveFrozenSet)
    assert s & test_set == actual
    assert actual.original_case == exp_case



# -*- coding: utf-8 -*-
"""Contains tests for the set_nmmpmat routine used for modifying the
   density matrix for LDA+U calculations."""
import os
import pytest
import numpy as np

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_INPXML_LDAU_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/GaAsMultiUForceXML/files/inp.xml')
TEST_NMMPMAT_PATH = os.path.join(FILE_PATH, 'files/fleur/input_nmmpmat.txt')


def test_set_nmmpmat_nofile(load_inpxml, file_regression):
    """Test setting of nmmpmat with no initial nmmpmat file given"""
    from masci_tools.util.xml.xml_setters_nmmpmat import set_nmmpmat

    xmltree, schema_dict = load_inpxml(TEST_INPXML_LDAU_PATH)

    nmmp_lines = None
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             species_name='Ga-1',
                             orbital=2,
                             spin=1,
                             state_occupations=[1, 2, 3, 4, 5])
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             'As-2',
                             orbital=1,
                             spin=1,
                             denmat=[[1, -2, 3], [4, -5, 6], [7, -8, 9]])

    file_regression.check(prepare_for_file_dump(nmmp_lines))


def test_set_nmmpmat_file(load_inpxml, file_regression):
    """Test setting of nmmpmat with initial nmmpmat file given"""
    from masci_tools.util.xml.xml_setters_nmmpmat import set_nmmpmat

    xmltree, schema_dict = load_inpxml(TEST_INPXML_LDAU_PATH)

    with open(TEST_NMMPMAT_PATH, mode='r') as nmmpfile:
        nmmp_lines = nmmpfile.read().split('\n')

    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             species_name='Ga-1',
                             orbital=2,
                             spin=1,
                             state_occupations=[1, 2, 3, 4, 5])
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             'As-2',
                             orbital=1,
                             spin=1,
                             denmat=[[1, -2, 3], [4, -5, 6], [7, -8, 9]])

    file_regression.check(prepare_for_file_dump(nmmp_lines))


def test_set_nmmpmat_file_get_wigner_matrix(load_inpxml, file_regression):
    """Test get_wigner_matrix by calling set_nmmpmat_file with theta, or phi != None"""
    from masci_tools.util.xml.xml_setters_nmmpmat import set_nmmpmat

    xmltree, schema_dict = load_inpxml(TEST_INPXML_LDAU_PATH)

    nmmp_lines = None
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             species_name='Ga-1',
                             orbital=1,
                             spin=1,
                             state_occupations=[1, 0, 1],
                             theta=np.pi / 2.0)
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             'As-2',
                             orbital=1,
                             spin=1,
                             denmat=[[1, 0, 1], [0, 0, 0], [1, 0, 1]],
                             phi=np.pi / 4.0,
                             theta=np.pi / 2.0)

    file_regression.check(prepare_for_file_dump(nmmp_lines))


def test_rotate_nmmpmat(load_inpxml, file_regression):
    """Test get_wigner_matrix by calling set_nmmpmat_file with theta, or phi != None"""
    from masci_tools.util.xml.xml_setters_nmmpmat import set_nmmpmat, rotate_nmmpmat

    xmltree, schema_dict = load_inpxml(TEST_INPXML_LDAU_PATH)

    nmmp_lines = None
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             species_name='Ga-1',
                             orbital=1,
                             spin=1,
                             state_occupations=[1, 0, 1])
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             'As-2',
                             orbital=1,
                             spin=1,
                             denmat=[[1, 0, 1], [0, 0, 0], [1, 0, 1]])

    nmmp_lines = rotate_nmmpmat(xmltree, nmmp_lines, schema_dict, 'Ga-1', orbital=1, phi=0.0, theta=np.pi / 2.0)
    nmmp_lines = rotate_nmmpmat(xmltree, nmmp_lines, schema_dict, 'As-2', orbital=1, theta=np.pi / 2.0, phi=np.pi / 4.0)

    file_regression.check(prepare_for_file_dump(nmmp_lines))


def test_validate_nmmpmat(load_inpxml):
    """Test validation method of nmmpmat file together with inp.xml file"""
    from masci_tools.util.xml.xml_setters_nmmpmat import set_nmmpmat, validate_nmmpmat

    xmltree, schema_dict = load_inpxml(TEST_INPXML_LDAU_PATH)

    with open(TEST_NMMPMAT_PATH, mode='r') as nmmpfile:
        nmmp_lines_orig = nmmpfile.read().split('\n')

    validate_nmmpmat(xmltree, nmmp_lines_orig, schema_dict)  #should not raise

    #Test number of lines error
    nmmp_lines = nmmp_lines_orig.copy()
    nmmp_lines.append('0.0')
    with pytest.raises(ValueError):
        validate_nmmpmat(xmltree, nmmp_lines, schema_dict)
    nmmp_lines.remove('0.0')

    #Test invalid diagonal element error
    nmmp_lines = nmmp_lines_orig.copy()
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             species_name='Ga-1',
                             orbital=2,
                             spin=1,
                             state_occupations=[1, 2, 3, 4, 5])
    nmmp_lines = set_nmmpmat(xmltree,
                             nmmp_lines,
                             schema_dict,
                             'As-2',
                             orbital=1,
                             spin=1,
                             denmat=[[1, -2, 3], [4, -5, 6], [7, -8, 9]])
    with pytest.raises(ValueError):
        validate_nmmpmat(xmltree, nmmp_lines, schema_dict)

    #Test invalid outsied value error
    nmmp_lines = nmmp_lines_orig.copy()
    nmmp_lines[
        0] = '     0.0000000000000     9.0000000000000     0.0000000000000     0.0000000000000     0.0000000000000     0.0000000000000     0.0000000000000'

    with pytest.raises(ValueError):
        validate_nmmpmat(xmltree, nmmp_lines, schema_dict)


def prepare_for_file_dump(file_lines):
    """
    Join lines together with linebreaks and remove negative zeros
    """
    return '\n'.join([line.replace('-0.0000000000000', ' 0.0000000000000') for line in file_lines])
